from pretrainedmodels.models.senet import SENet
from pretrainedmodels.models.senet import SEBottleneck
from pretrainedmodels.models.senet import SEResNetBottleneck
from pretrainedmodels.models.senet import SEResNeXtBottleneck
from pretrainedmodels.models.senet import pretrained_settings


class SENetEncoder(SENet):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pretrained = False

        del self.last_linear
        del self.avg_pool

    def forward(self, x):
        for module in self.layer0[:-1]:
            x = module(x)

        x0 = x
        x = self.layer0[-1](x)
        x1 = self.layer1(x)
        x2 = self.layer2(x1)
        x3 = self.layer3(x2)
        x4 = self.layer4(x3)

        features = [x4, x3, x2, x1, x0]
        return features

    def load_state_dict(self, state_dict, **kwargs):
        state_dict.pop('last_linear.bias')
        state_dict.pop('last_linear.weight')
        super().load_state_dict(state_dict, **kwargs)


senet_encoders = {
    'senet154': {
        'encoder': SENetEncoder,
        'pretrained_settings': pretrained_settings['senet154'],
        'out_shapes': (2048, 1024, 512, 256, 128),
        'params': {
            'block': SEBottleneck,
            'dropout_p': 0.2,
            'groups': 64,
            'layers': [3, 8, 36, 3],
            'num_classes': 1000,
            'reduction': 16
        },
    },

    'se_resnet50': {
        'encoder': SENetEncoder,
        'pretrained_settings': pretrained_settings['se_resnet50'],
        'out_shapes': (2048, 1024, 512, 256, 64),
        'params': {
            'block': SEResNetBottleneck,
            'layers': [3, 4, 6, 3],
            'downsample_kernel_size': 1,
            'downsample_padding': 0,
            'dropout_p': None,
            'groups': 1,
            'inplanes': 64,
            'input_3x3': False,
            'num_classes': 1000,
            'reduction': 16
        },
    },

    'se_resnet101': {
        'encoder': SENetEncoder,
        'pretrained_settings': pretrained_settings['se_resnet101'],
        'out_shapes': (2048, 1024, 512, 256, 64),
        'params': {
            'block': SEResNetBottleneck,
            'layers': [3, 4, 23, 3],
            'downsample_kernel_size': 1,
            'downsample_padding': 0,
            'dropout_p': None,
            'groups': 1,
            'inplanes': 64,
            'input_3x3': False,
            'num_classes': 1000,
            'reduction': 16
        },
    },

    'se_resnet152': {
        'encoder': SENetEncoder,
        'pretrained_settings': pretrained_settings['se_resnet152'],
        'out_shapes': (2048, 1024, 512, 256, 64),
        'params': {
            'block': SEResNetBottleneck,
            'layers': [3, 8, 36, 3],
            'downsample_kernel_size': 1,
            'downsample_padding': 0,
            'dropout_p': None,
            'groups': 1,
            'inplanes': 64,
            'input_3x3': False,
            'num_classes': 1000,
            'reduction': 16
        },
    },

    'se_resnext50_32x4d': {
        'encoder': SENetEncoder,
        'pretrained_settings': pretrained_settings['se_resnext50_32x4d'],
        'out_shapes': (2048, 1024, 512, 256, 64),
        'params': {
            'block': SEResNeXtBottleneck,
            'layers': [3, 4, 6, 3],
            'downsample_kernel_size': 1,
            'downsample_padding': 0,
            'dropout_p': None,
            'groups': 32,
            'inplanes': 64,
            'input_3x3': False,
            'num_classes': 1000,
            'reduction': 16
        },
    },

    'se_resnext101_32x4d': {
        'encoder': SENetEncoder,
        'pretrained_settings': pretrained_settings['se_resnext101_32x4d'],
        'out_shapes': (2048, 1024, 512, 256, 64),
        'params': {
            'block': SEResNeXtBottleneck,
            'layers': [3, 4, 23, 3],
            'downsample_kernel_size': 1,
            'downsample_padding': 0,
            'dropout_p': None,
            'groups': 32,
            'inplanes': 64,
            'input_3x3': False,
            'num_classes': 1000,
            'reduction': 16
        },
    },
}



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch.nn as nn

from .r2plus1_util import r2plus1_unit


class ResNeXt3DStemSinglePathway(nn.Module):
    """
    ResNe(X)t 3D basic stem module. Assume a single pathway.
    Performs spatiotemporal Convolution, BN, and Relu following by a
        spatiotemporal pooling.
    """

    def __init__(
            self,
            dim_in,
            dim_out,
            kernel,
            stride,
            padding,
            maxpool=True,
            inplace_relu=True,
            bn_eps=1e-5,
            bn_mmt=0.1,
    ):
        """
        The `__init__` method of any subclass should also contain these arguments.

        Args:
            dim_in (int): the channel dimension of the input. Normally 3 is used
                for rgb input
            dim_out (int): the output dimension of the convolution in the stem
                layer.
            kernel (list): the kernel size of the convolution in the stem layer.
                temporal kernel size, height kernel size, width kernel size in
                order.
            stride (list): the stride size of the convolution in the stem layer.
                temporal kernel stride, height kernel size, width kernel size in
                order.
            padding (int): the padding size of the convolution in the stem
                layer, temporal padding size, height padding size, width
                padding size in order.
            maxpool (bool): If true, perform max pooling.
            inplace_relu (bool): calculate the relu on the original input
                without allocating new memory.
            bn_eps (float): epsilon for batch norm.
            bn_mmt (float): momentum for batch norm. Noted that BN momentum in
                PyTorch = 1 - BN momentum in Caffe2.
        """
        super(ResNeXt3DStemSinglePathway, self).__init__()
        self.kernel = kernel
        self.stride = stride
        self.padding = padding
        self.inplace_relu = inplace_relu
        self.bn_eps = bn_eps
        self.bn_mmt = bn_mmt
        self.maxpool = maxpool

        # Construct the stem layer.
        self._construct_stem(dim_in, dim_out)

    def _construct_stem(self, dim_in, dim_out):
        self.conv = nn.Conv3d(
            dim_in,
            dim_out,
            self.kernel,
            stride=self.stride,
            padding=self.padding,
            bias=False,
        )
        self.bn = nn.BatchNorm3d(dim_out, eps=self.bn_eps,
                                 momentum=self.bn_mmt)
        self.relu = nn.ReLU(self.inplace_relu)
        if self.maxpool:
            self.pool_layer = nn.MaxPool3d(
                kernel_size=[1, 3, 3], stride=[1, 2, 2], padding=[0, 1, 1]
            )

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        if self.maxpool:
            x = self.pool_layer(x)
        return x


class R2Plus1DStemSinglePathway(ResNeXt3DStemSinglePathway):
    """
    R(2+1)D basic stem module. Assume a single pathway.
    Performs spatial convolution, temporal convolution, BN, and Relu following by a
        spatiotemporal pooling.
    """

    def __init__(
            self,
            dim_in,
            dim_out,
            kernel,
            stride,
            padding,
            maxpool=True,
            inplace_relu=True,
            bn_eps=1e-5,
            bn_mmt=0.1,
    ):
        """
        The `__init__` method of any subclass should also contain these arguments.

        Args:
            dim_in (int): the channel dimension of the input. Normally 3 is used
                for rgb input
            dim_out (int): the output dimension of the convolution in the stem
                layer.
            kernel (list): the kernel size of the convolution in the stem layer.
                temporal kernel size, height kernel size, width kernel size in
                order.
            stride (list): the stride size of the convolution in the stem layer.
                temporal kernel stride, height kernel size, width kernel size in
                order.
            padding (int): the padding size of the convolution in the stem
                layer, temporal padding size, height padding size, width
                padding size in order.
            maxpool (bool): If true, perform max pooling.
            inplace_relu (bool): calculate the relu on the original input
                without allocating new memory.
            bn_eps (float): epsilon for batch norm.
            bn_mmt (float): momentum for batch norm. Noted that BN momentum in
                PyTorch = 1 - BN momentum in Caffe2.
        """
        super(R2Plus1DStemSinglePathway, self).__init__(
            dim_in,
            dim_out,
            kernel,
            stride,
            padding,
            maxpool=maxpool,
            inplace_relu=inplace_relu,
            bn_eps=bn_eps,
            bn_mmt=bn_mmt,
        )

    def _construct_stem(self, dim_in, dim_out):
        assert (
                self.stride[1] == self.stride[2]
        ), "Only support identical height stride and width stride"
        self.conv = r2plus1_unit(
            dim_in,
            dim_out,
            self.stride[0],  # temporal_stride
            self.stride[1],  # spatial_stride
            1,  # groups
            self.inplace_relu,
            self.bn_eps,
            self.bn_mmt,
            dim_mid=45,  # hard-coded middle channels
        )
        self.bn = nn.BatchNorm3d(dim_out, eps=self.bn_eps,
                                 momentum=self.bn_mmt)
        self.relu = nn.ReLU(self.inplace_relu)
        if self.maxpool:
            self.pool_layer = nn.MaxPool3d(
                kernel_size=[1, 3, 3], stride=[1, 2, 2], padding=[0, 1, 1]
            )


class ResNeXt3DStemMultiPathway(nn.Module):
    """
    Video 3D stem module. Provides stem operations of Conv, BN, ReLU, MaxPool
    on input data tensor for one or multiple pathways.
    """

    def __init__(
            self,
            dim_in,
            dim_out,
            kernel,
            stride,
            padding,
            inplace_relu=True,
            bn_eps=1e-5,
            bn_mmt=0.1,
            maxpool=(True,),
    ):
        """
        The `__init__` method of any subclass should also contain these
        arguments. List size of 1 for single pathway models (C2D, I3D, SlowOnly
        and etc), list size of 2 for two pathway models (SlowFast).

        Args:
            dim_in (list): the list of channel dimensions of the inputs.
            dim_out (list): the output dimension of the convolution in the stem
                layer.
            kernel (list): the kernels' size of the convolutions in the stem
                layers. Temporal kernel size, height kernel size, width kernel
                size in order.
            stride (list): the stride sizes of the convolutions in the stem
                layer. Temporal kernel stride, height kernel size, width kernel
                size in order.
            padding (list): the paddings' sizes of the convolutions in the stem
                layer. Temporal padding size, height padding size, width padding
                size in order.
            inplace_relu (bool): calculate the relu on the original input
                without allocating new memory.
            bn_eps (float): epsilon for batch norm.
            bn_mmt (float): momentum for batch norm. Noted that BN momentum in
                PyTorch = 1 - BN momentum in Caffe2.
            maxpool (iterable): At training time, when crop size is 224 x 224, do max
                pooling. When crop size is 112 x 112, skip max pooling.
                Default value is a (True,)
        """
        super(ResNeXt3DStemMultiPathway, self).__init__()

        assert (
                len({len(dim_in), len(dim_out), len(kernel), len(stride),
                     len(padding)})
                == 1
        ), "Input pathway dimensions are not consistent."
        self.num_pathways = len(dim_in)
        self.kernel = kernel
        self.stride = stride
        self.padding = padding
        self.inplace_relu = inplace_relu
        self.bn_eps = bn_eps
        self.bn_mmt = bn_mmt
        self.maxpool = maxpool

        # Construct the stem layer.
        self._construct_stem(dim_in, dim_out)

    def _construct_stem(self, dim_in, dim_out):
        assert type(dim_in) == list
        assert all(dim > 0 for dim in dim_in)
        assert type(dim_out) == list
        assert all(dim > 0 for dim in dim_out)

        self.blocks = {}
        for p in range(len(dim_in)):
            stem = ResNeXt3DStemSinglePathway(
                dim_in[p],
                dim_out[p],
                self.kernel[p],
                self.stride[p],
                self.padding[p],
                inplace_relu=self.inplace_relu,
                bn_eps=self.bn_eps,
                bn_mmt=self.bn_mmt,
                maxpool=self.maxpool[p],
            )
            stem_name = self._stem_name(p)
            self.add_module(stem_name, stem)
            self.blocks[stem_name] = stem

    def _stem_name(self, path_idx):
        return "stem-path{}".format(path_idx)

    def forward(self, x):
        assert (
                len(x) == self.num_pathways
        ), "Input tensor does not contain {} pathway".format(self.num_pathways)
        for p in range(len(x)):
            stem_name = self._stem_name(p)
            x[p] = self.blocks[stem_name](x[p])
        return x


class R2Plus1DStemMultiPathway(ResNeXt3DStemMultiPathway):
    """
    Video R(2+1)D stem module. Provides stem operations of Conv, BN, ReLU, MaxPool
    on input data tensor for one or multiple pathways.
    """

    def __init__(
            self,
            dim_in,
            dim_out,
            kernel,
            stride,
            padding,
            inplace_relu=True,
            bn_eps=1e-5,
            bn_mmt=0.1,
            maxpool=(True,),
    ):
        """
        The `__init__` method of any subclass should also contain these
        arguments. List size of 1 for single pathway models (C2D, I3D, SlowOnly
        and etc), list size of 2 for two pathway models (SlowFast).

        Args:
            dim_in (list): the list of channel dimensions of the inputs.
            dim_out (list): the output dimension of the convolution in the stem
                layer.
            kernel (list): the kernels' size of the convolutions in the stem
                layers. Temporal kernel size, height kernel size, width kernel
                size in order.
            stride (list): the stride sizes of the convolutions in the stem
                layer. Temporal kernel stride, height kernel size, width kernel
                size in order.
            padding (list): the paddings' sizes of the convolutions in the stem
                layer. Temporal padding size, height padding size, width padding
                size in order.
            inplace_relu (bool): calculate the relu on the original input
                without allocating new memory.
            bn_eps (float): epsilon for batch norm.
            bn_mmt (float): momentum for batch norm. Noted that BN momentum in
                PyTorch = 1 - BN momentum in Caffe2.
            maxpool (iterable): At training time, when crop size is 224 x 224, do max
                pooling. When crop size is 112 x 112, skip max pooling.
                Default value is a (True,)
        """
        super(R2Plus1DStemMultiPathway, self).__init__(
            dim_in,
            dim_out,
            kernel,
            stride,
            padding,
            inplace_relu=inplace_relu,
            bn_eps=bn_eps,
            bn_mmt=bn_mmt,
            maxpool=maxpool,
        )

    def _construct_stem(self, dim_in, dim_out):
        assert type(dim_in) == list
        assert all(dim > 0 for dim in dim_in)
        assert type(dim_out) == list
        assert all(dim > 0 for dim in dim_out)

        self.blocks = {}
        for p in range(len(dim_in)):
            stem = R2Plus1DStemSinglePathway(
                dim_in[p],
                dim_out[p],
                self.kernel[p],
                self.stride[p],
                self.padding[p],
                inplace_relu=self.inplace_relu,
                bn_eps=self.bn_eps,
                bn_mmt=self.bn_mmt,
                maxpool=self.maxpool[p],
            )
            stem_name = self._stem_name(p)
            self.add_module(stem_name, stem)
            self.blocks[stem_name] = stem


class ResNeXt3DStem(nn.Module):
    def __init__(
            self, temporal_kernel, spatial_kernel, input_planes, stem_planes,
            maxpool
    ):
        super(ResNeXt3DStem, self).__init__()
        self._construct_stem(
            temporal_kernel, spatial_kernel, input_planes, stem_planes, maxpool
        )

    def _construct_stem(
            self, temporal_kernel, spatial_kernel, input_planes, stem_planes,
            maxpool
    ):
        self.stem = ResNeXt3DStemMultiPathway(
            [input_planes],
            [stem_planes],
            [[temporal_kernel, spatial_kernel, spatial_kernel]],
            [[1, 2, 2]],  # stride
            [
                [temporal_kernel // 2, spatial_kernel // 2,
                 spatial_kernel // 2]
            ],  # padding
            maxpool=[maxpool],
        )

    def forward(self, x):
        return self.stem(x)


class R2Plus1DStem(ResNeXt3DStem):
    def __init__(
            self, temporal_kernel, spatial_kernel, input_planes, stem_planes,
            maxpool
    ):
        super(R2Plus1DStem, self).__init__(
            temporal_kernel, spatial_kernel, input_planes, stem_planes, maxpool
        )

    def _construct_stem(
            self, temporal_kernel, spatial_kernel, input_planes, stem_planes,
            maxpool
    ):
        self.stem = R2Plus1DStemMultiPathway(
            [input_planes],
            [stem_planes],
            [[temporal_kernel, spatial_kernel, spatial_kernel]],
            [[1, 2, 2]],  # stride
            [
                [temporal_kernel // 2, spatial_kernel // 2,
                 spatial_kernel // 2]
            ],  # padding
            maxpool=[maxpool],
        )



import os
from os.path import join
import shutil

from .__version__ import __version__  # noqa: F401

ROOT_FOLDER = os.path.abspath(
    os.path.expanduser(os.getenv('ROOT_FOLDER', '~/mlcomp')))

test_worker = os.getenv('PYTEST_XDIST_WORKER')
if test_worker:
    ROOT_FOLDER = join(ROOT_FOLDER, 'tests', test_worker)
    shutil.rmtree(ROOT_FOLDER, ignore_errors=True)

DATA_FOLDER = join(ROOT_FOLDER, 'data')
MODEL_FOLDER = join(ROOT_FOLDER, 'models')
TASK_FOLDER = join(ROOT_FOLDER, 'tasks')
LOG_FOLDER = join(ROOT_FOLDER, 'logs')
CONFIG_FOLDER = join(ROOT_FOLDER, 'configs')
DB_FOLDER = join(ROOT_FOLDER, 'db')
REPORT_FOLDER = join(ROOT_FOLDER, 'report')
TMP_FOLDER = join(ROOT_FOLDER, 'tmp')

os.makedirs(ROOT_FOLDER, exist_ok=True)
os.makedirs(DATA_FOLDER, exist_ok=True)
os.makedirs(MODEL_FOLDER, exist_ok=True)
os.makedirs(TASK_FOLDER, exist_ok=True)
os.makedirs(LOG_FOLDER, exist_ok=True)
os.makedirs(CONFIG_FOLDER, exist_ok=True)
os.makedirs(DB_FOLDER, exist_ok=True)
os.makedirs(REPORT_FOLDER, exist_ok=True)
os.makedirs(TMP_FOLDER, exist_ok=True)

# copy conf files if they do not exist

folder = os.path.dirname(__file__)
docker_folder = join(folder, 'docker')

for name in os.listdir(docker_folder):
    file = join(docker_folder, name)
    target_file = join(CONFIG_FOLDER, name)
    if not os.path.exists(target_file):
        shutil.copy(file, target_file)

# exporting environment variables
env_file = join(CONFIG_FOLDER, '.env')

with open(env_file) as f:
    for l in f.readlines():
        k, v = l.strip().split('=')
        os.environ[k] = v

# extra env
if os.getenv('ENV'):
    extra_env_file = join(CONFIG_FOLDER, os.getenv('ENV', '')+'.env')
    extra_env = open(extra_env_file).readlines()
    for p in extra_env:
        if '=' not in p:
            continue
        k, v = p.strip().split('=')
        os.environ[k] = v

# contour env file
CONTOUR_FILE = join(CONFIG_FOLDER, 'contour.yml')

# for debugging
os.environ['PYTHONPATH'] = '.'

MASTER_PORT_RANGE = list(map(int, os.getenv('MASTER_PORT_RANGE').split('-')))
FILE_SYNC_INTERVAL = int(os.getenv('FILE_SYNC_INTERVAL', '0'))
WORKER_USAGE_INTERVAL = int(os.getenv('WORKER_USAGE_INTERVAL', '10'))
INSTALL_DEPENDENCIES = os.getenv('INSTALL_DEPENDENCIES') == 'True'

REDIS_HOST = os.getenv('REDIS_HOST')
REDIS_PASSWORD = os.getenv('REDIS_PASSWORD')
REDIS_PORT = os.getenv('REDIS_PORT')

TOKEN = os.getenv('TOKEN')
DOCKER_IMG = os.getenv('DOCKER_IMG', 'default')
WEB_HOST = os.getenv('WEB_HOST')
WEB_PORT = int(os.getenv('WEB_PORT'))
WORKER_INDEX = os.getenv('WORKER_INDEX', -1)

CONSOLE_LOG_LEVEL = os.getenv('CONSOLE_LOG_LEVEL', 'DEBUG')
DB_LOG_LEVEL = os.getenv('DB_LOG_LEVEL', 'DEBUG')
FILE_LOG_LEVEL = os.getenv('FILE_LOG_LEVEL', 'INFO')
LOG_NAME = os.getenv('LOG_NAME', 'log')
SYNC_WITH_THIS_COMPUTER = os.getenv('SYNC_WITH_THIS_COMPUTER') == 'True'
CAN_PROCESS_TASKS = os.getenv('CAN_PROCESS_TASKS') == 'True'

DB_TYPE = os.getenv('DB_TYPE')
if DB_TYPE == 'POSTGRESQL':
    DATABASE = {
        'dbname': os.getenv('POSTGRES_DB'),
        'user': os.getenv('POSTGRES_USER'),
        'password': os.getenv('POSTGRES_PASSWORD'),
        'host': os.getenv('POSTGRES_HOST'),
        'port': int(os.getenv('POSTGRES_PORT')),
    }

    SA_CONNECTION_STRING = f"postgresql+psycopg2://{DATABASE['user']}:" \
                           f"{DATABASE['password']}@{DATABASE['host']}:" \
                           f"{DATABASE['port']}/{DATABASE['dbname']}"
elif DB_TYPE == 'SQLITE':
    SA_CONNECTION_STRING = f'sqlite:///{DB_FOLDER}/sqlite3.sqlite'
else:
    raise Exception(f'Unknown DB_TYPE = {DB_TYPE}')

FLASK_ENV = os.getenv('FLASK_ENV')
DOCKER_MAIN = os.getenv('DOCKER_MAIN', 'True') == 'True'

IP = os.getenv('IP')
PORT = int(os.getenv('PORT'))

__all__ = [
    'ROOT_FOLDER', 'DATA_FOLDER', 'MODEL_FOLDER', 'TASK_FOLDER', 'LOG_FOLDER',
    'CONFIG_FOLDER', 'DB_FOLDER', 'MASTER_PORT_RANGE', 'REDIS_HOST',
    'REDIS_PASSWORD', 'REDIS_PORT', 'TOKEN', 'DOCKER_IMG', 'WEB_HOST',
    'WEB_PORT', 'WORKER_INDEX', 'CONSOLE_LOG_LEVEL', 'DB_LOG_LEVEL',
    'FILE_LOG_LEVEL', 'DB_TYPE', 'SA_CONNECTION_STRING', 'FLASK_ENV',
    'DOCKER_MAIN', 'IP', 'PORT', 'LOG_NAME', 'WORKER_USAGE_INTERVAL',
    'FILE_SYNC_INTERVAL', 'INSTALL_DEPENDENCIES', 'SYNC_WITH_THIS_COMPUTER',
    'CAN_PROCESS_TASKS', 'TMP_FOLDER', 'CONTOUR_FILE', 'REPORT_FOLDER'
]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch.nn as nn

from .r2plus1_util import r2plus1_unit


class BasicTransformation(nn.Module):
    """
    Basic transformation: 3x3x3 group conv, 3x3x3 group conv
    """

    def __init__(
        self,
        dim_in,
        dim_out,
        temporal_stride,
        spatial_stride,
        groups,
        inplace_relu=True,
        bn_eps=1e-5,
        bn_mmt=0.1,
        **kwargs
    ):
        """
        Args:
            dim_in (int): the channel dimensions of the input.
            dim_out (int): the channel dimension of the output.
            temporal_stride (int): the temporal stride of the bottleneck.
            spatial_stride (int): the spatial_stride of the bottleneck.
            groups (int): number of groups for the convolution.
            inplace_relu (bool): calculate the relu on the original input
                without allocating new memory.
            bn_eps (float): epsilon for batch norm.
            bn_mmt (float): momentum for batch norm. Noted that BN momentum in
                PyTorch = 1 - BN momentum in Caffe2.
        """
        super(BasicTransformation, self).__init__()
        self._construct_model(
            dim_in,
            dim_out,
            temporal_stride,
            spatial_stride,
            groups,
            inplace_relu,
            bn_eps,
            bn_mmt,
        )

    def _construct_model(
        self,
        dim_in,
        dim_out,
        temporal_stride,
        spatial_stride,
        groups,
        inplace_relu,
        bn_eps,
        bn_mmt,
    ):
        # 3x3x3 group conv, BN, ReLU.
        branch2a = nn.Conv3d(
            dim_in,
            dim_out,
            [3, 3, 3],  # kernel
            stride=[temporal_stride, spatial_stride, spatial_stride],
            padding=[1, 1, 1],
            groups=groups,
            bias=False,
        )
        branch2a_bn = nn.BatchNorm3d(dim_out, eps=bn_eps, momentum=bn_mmt)
        branch2a_relu = nn.ReLU(inplace=inplace_relu)
        # 3x3x3 group conv, BN, ReLU.
        branch2b = nn.Conv3d(
            dim_out,
            dim_out,
            [3, 3, 3],  # kernel
            stride=[1, 1, 1],
            padding=[1, 1, 1],
            groups=groups,
            bias=False,
        )
        branch2b_bn = nn.BatchNorm3d(dim_out, eps=bn_eps, momentum=bn_mmt)
        branch2b_bn.final_transform_op = True

        self.transform = nn.Sequential(
            branch2a, branch2a_bn, branch2a_relu, branch2b, branch2b_bn
        )

    def forward(self, x):
        return self.transform(x)


class BasicR2Plus1DTransformation(BasicTransformation):
    """
    Basic transformation: 3x3x3 group conv, 3x3x3 group conv
    """

    def __init__(
        self,
        dim_in,
        dim_out,
        temporal_stride,
        spatial_stride,
        groups,
        inplace_relu=True,
        bn_eps=1e-5,
        bn_mmt=0.1,
        **kwargs
    ):
        """
        Args:
            dim_in (int): the channel dimensions of the input.
            dim_out (int): the channel dimension of the output.
            temporal_stride (int): the temporal stride of the bottleneck.
            spatial_stride (int): the spatial_stride of the bottleneck.
            groups (int): number of groups for the convolution.
            inplace_relu (bool): calculate the relu on the original input
                without allocating new memory.
            bn_eps (float): epsilon for batch norm.
            bn_mmt (float): momentum for batch norm. Noted that BN momentum in
                PyTorch = 1 - BN momentum in Caffe2.
        """
        super(BasicR2Plus1DTransformation, self).__init__(
            dim_in,
            dim_out,
            temporal_stride,
            spatial_stride,
            groups,
            inplace_relu=inplace_relu,
            bn_eps=bn_eps,
            bn_mmt=bn_mmt,
        )

    def _construct_model(
        self,
        dim_in,
        dim_out,
        temporal_stride,
        spatial_stride,
        groups,
        inplace_relu,
        bn_eps,
        bn_mmt,
    ):
        # Implementation of R(2+1)D operation <https://arxiv.org/abs/1711.11248>.
        # decompose the original 3D conv into one 2D spatial conv and one
        # 1D temporal conv
        branch2a = r2plus1_unit(
            dim_in,
            dim_out,
            temporal_stride,
            spatial_stride,
            groups,
            inplace_relu,
            bn_eps,
            bn_mmt,
        )
        branch2a_bn = nn.BatchNorm3d(dim_out, eps=bn_eps, momentum=bn_mmt)
        branch2a_relu = nn.ReLU(inplace=inplace_relu)

        branch2b = r2plus1_unit(
            dim_out,
            dim_out,
            1,  # temporal_stride
            1,  # spatial_stride
            groups,
            inplace_relu,
            bn_eps,
            bn_mmt,
        )
        branch2b_bn = nn.BatchNorm3d(dim_out, eps=bn_eps, momentum=bn_mmt)
        branch2b_bn.final_transform_op = True

        self.transform = nn.Sequential(
            branch2a, branch2a_bn, branch2a_relu, branch2b, branch2b_bn
        )


class PostactivatedBottleneckTransformation(nn.Module):
    """
    Bottleneck transformation: Tx1x1, 1x3x3, 1x1x1, where T is the size of
        temporal kernel.
    """

    def __init__(
        self,
        dim_in,
        dim_out,
        temporal_stride,
        spatial_stride,
        num_groups,
        dim_inner,
        temporal_kernel_size=3,
        temporal_conv_1x1=True,
        spatial_stride_1x1=False,
        inplace_relu=True,
        bn_eps=1e-5,
        bn_mmt=0.1,
        **kwargs
    ):
        """
        Args:
            dim_in (int): the channel dimensions of the input.
            dim_out (int): the channel dimension of the output.
            temporal_kernel_size (int): the temporal kernel sizes of the middle
                convolution in the bottleneck.
            temporal_conv_1x1 (bool): if True, do temporal convolution in the fist
                1x1 Conv3d. Otherwise, do it in the second 3x3 Conv3d
            temporal_stride (int): the temporal stride of the bottleneck.
            spatial_stride (int): the spatial_stride of the bottleneck.
            num_groups (int): number of groups for the convolution.
            dim_inner (int): the inner dimension of the block.
                is for standard ResNet like networks, and num_groups>1 is for
                ResNeXt like networks.
            spatial_stride_1x1 (bool): if True, apply spatial_stride to 1x1 conv.
            inplace_relu (bool): calculate the relu on the original input
                without allocating new memory.
            bn_eps (float): epsilon for batch norm.
            bn_mmt (float): momentum for batch norm. Noted that BN momentum in
                PyTorch = 1 - BN momentum in Caffe2.
        """
        super(PostactivatedBottleneckTransformation, self).__init__()
        (temporal_kernel_size_1x1, temporal_kernel_size_3x3) = (
            (temporal_kernel_size, 1)
            if temporal_conv_1x1
            else (1, temporal_kernel_size)
        )
        # MSRA -> stride=2 is on 1x1; TH/C2 -> stride=2 is on 3x3.
        (str1x1, str3x3) = (
            (spatial_stride, 1) if spatial_stride_1x1 else (1, spatial_stride)
        )
        # Tx1x1 conv, BN, ReLU.
        self.branch2a = nn.Conv3d(
            dim_in,
            dim_inner,
            kernel_size=[temporal_kernel_size_1x1, 1, 1],
            stride=[1, str1x1, str1x1],
            padding=[temporal_kernel_size_1x1 // 2, 0, 0],
            bias=False,
        )
        self.branch2a_bn = nn.BatchNorm3d(dim_inner, eps=bn_eps, momentum=bn_mmt)
        self.branch2a_relu = nn.ReLU(inplace=inplace_relu)
        # Tx3x3 group conv, BN, ReLU.
        self.branch2b = nn.Conv3d(
            dim_inner,
            dim_inner,
            [temporal_kernel_size_3x3, 3, 3],
            stride=[temporal_stride, str3x3, str3x3],
            padding=[temporal_kernel_size_3x3 // 2, 1, 1],
            groups=num_groups,
            bias=False,
        )
        self.branch2b_bn = nn.BatchNorm3d(dim_inner, eps=bn_eps, momentum=bn_mmt)
        self.branch2b_relu = nn.ReLU(inplace=inplace_relu)
        # 1x1x1 conv, BN.
        self.branch2c = nn.Conv3d(
            dim_inner,
            dim_out,
            kernel_size=[1, 1, 1],
            stride=[1, 1, 1],
            padding=[0, 0, 0],
            bias=False,
        )
        self.branch2c_bn = nn.BatchNorm3d(dim_out, eps=bn_eps, momentum=bn_mmt)
        self.branch2c_bn.final_transform_op = True

    def forward(self, x):
        # Explicitly forward every layer.
        # Branch2a.
        x = self.branch2a(x)
        x = self.branch2a_bn(x)
        x = self.branch2a_relu(x)

        # Branch2b.
        x = self.branch2b(x)
        x = self.branch2b_bn(x)
        x = self.branch2b_relu(x)

        # Branch2c
        x = self.branch2c(x)
        x = self.branch2c_bn(x)
        return x


class PreactivatedBottleneckTransformation(nn.Module):
    """
    Bottleneck transformation with pre-activation, which includes BatchNorm3D
        and ReLu. Conv3D kernsl are Tx1x1, 1x3x3, 1x1x1, where T is the size of
        temporal kernel (https://arxiv.org/abs/1603.05027).
    """

    def __init__(
        self,
        dim_in,
        dim_out,
        temporal_stride,
        spatial_stride,
        num_groups,
        dim_inner,
        temporal_kernel_size=3,
        temporal_conv_1x1=True,
        spatial_stride_1x1=False,
        inplace_relu=True,
        bn_eps=1e-5,
        bn_mmt=0.1,
        disable_pre_activation=False,
        **kwargs
    ):
        """
        Args:
            dim_in (int): the channel dimensions of the input.
            dim_out (int): the channel dimension of the output.
            temporal_kernel_size (int): the temporal kernel sizes of the middle
                convolution in the bottleneck.
            temporal_conv_1x1 (bool): if True, do temporal convolution in the fist
                1x1 Conv3d. Otherwise, do it in the second 3x3 Conv3d
            temporal_stride (int): the temporal stride of the bottleneck.
            spatial_stride (int): the spatial_stride of the bottleneck.
            num_groups (int): number of groups for the convolution.
            dim_inner (int): the inner dimension of the block.
                is for standard ResNet like networks, and num_groups>1 is for
                ResNeXt like networks.
            spatial_stride_1x1 (bool): if True, apply spatial_stride to 1x1 conv.
            inplace_relu (bool): calculate the relu on the original input
                without allocating new memory.
            bn_eps (float): epsilon for batch norm.
            bn_mmt (float): momentum for batch norm. Noted that BN momentum in
                PyTorch = 1 - BN momentum in Caffe2.
            disable_pre_activation (bool): If true, disable pre activation,
                including BatchNorm3D and ReLU.
        """
        super(PreactivatedBottleneckTransformation, self).__init__()
        (temporal_kernel_size_1x1, temporal_kernel_size_3x3) = (
            (temporal_kernel_size, 1)
            if temporal_conv_1x1
            else (1, temporal_kernel_size)
        )
        (str1x1, str3x3) = (
            (spatial_stride, 1) if spatial_stride_1x1 else (1, spatial_stride)
        )

        self.disable_pre_activation = disable_pre_activation
        if not disable_pre_activation:
            self.branch2a_bn = nn.BatchNorm3d(dim_in, eps=bn_eps, momentum=bn_mmt)
            self.branch2a_relu = nn.ReLU(inplace=inplace_relu)

        self.branch2a = nn.Conv3d(
            dim_in,
            dim_inner,
            kernel_size=[temporal_kernel_size_1x1, 1, 1],
            stride=[1, str1x1, str1x1],
            padding=[temporal_kernel_size_1x1 // 2, 0, 0],
            bias=False,
        )
        # Tx3x3 group conv, BN, ReLU.
        self.branch2b_bn = nn.BatchNorm3d(dim_inner, eps=bn_eps, momentum=bn_mmt)
        self.branch2b_relu = nn.ReLU(inplace=inplace_relu)
        self.branch2b = nn.Conv3d(
            dim_inner,
            dim_inner,
            [temporal_kernel_size_3x3, 3, 3],
            stride=[temporal_stride, str3x3, str3x3],
            padding=[temporal_kernel_size_3x3 // 2, 1, 1],
            groups=num_groups,
            bias=False,
        )
        # 1x1x1 conv, BN.
        self.branch2c_bn = nn.BatchNorm3d(dim_inner, eps=bn_eps, momentum=bn_mmt)
        self.branch2c_relu = nn.ReLU(inplace=inplace_relu)
        self.branch2c = nn.Conv3d(
            dim_inner,
            dim_out,
            kernel_size=[1, 1, 1],
            stride=[1, 1, 1],
            padding=[0, 0, 0],
            bias=False,
        )
        self.branch2c.final_transform_op = True

    def forward(self, x):
        # Branch2a
        if not self.disable_pre_activation:
            x = self.branch2a_bn(x)
            x = self.branch2a_relu(x)
        x = self.branch2a(x)
        # Branch2b
        x = self.branch2b_bn(x)
        x = self.branch2b_relu(x)
        x = self.branch2b(x)
        # Branch2c
        x = self.branch2c_bn(x)
        x = self.branch2c_relu(x)
        x = self.branch2c(x)
        return x


residual_transformations = {
    "basic_r2plus1d_transformation": BasicR2Plus1DTransformation,
    "basic_transformation": BasicTransformation,
    "postactivated_bottleneck_transformation": PostactivatedBottleneckTransformation,
    "preactivated_bottleneck_transformation": PreactivatedBottleneckTransformation,
    # For more types of residual transformations, add them below
}


class PostactivatedShortcutTransformation(nn.Module):
    """
    Skip connection used in ResNet3D model.
    """

    def __init__(
        self,
        dim_in,
        dim_out,
        temporal_stride,
        spatial_stride,
        bn_eps=1e-5,
        bn_mmt=0.1,
        **kwargs
    ):
        super(PostactivatedShortcutTransformation, self).__init__()
        # Use skip connection with projection if dim or spatial/temporal res change.
        assert (dim_in != dim_out) or (spatial_stride != 1) or (temporal_stride != 1)
        self.branch1 = nn.Conv3d(
            dim_in,
            dim_out,
            kernel_size=1,
            stride=[temporal_stride, spatial_stride, spatial_stride],
            padding=0,
            bias=False,
        )
        self.branch1_bn = nn.BatchNorm3d(dim_out, eps=bn_eps, momentum=bn_mmt)

    def forward(self, x):
        return self.branch1_bn(self.branch1(x))


class PreactivatedShortcutTransformation(nn.Module):
    """
    Skip connection with pre-activation, which includes BatchNorm3D and ReLU,
        in ResNet3D model (https://arxiv.org/abs/1603.05027).
    """

    def __init__(
        self,
        dim_in,
        dim_out,
        temporal_stride,
        spatial_stride,
        inplace_relu=True,
        bn_eps=1e-5,
        bn_mmt=0.1,
        disable_pre_activation=False,
        **kwargs
    ):
        super(PreactivatedShortcutTransformation, self).__init__()
        # Use skip connection with projection if dim or spatial/temporal res change.
        assert (dim_in != dim_out) or (spatial_stride != 1) or (temporal_stride != 1)
        if not disable_pre_activation:
            self.branch1_bn = nn.BatchNorm3d(dim_in, eps=bn_eps, momentum=bn_mmt)
            self.branch1_relu = nn.ReLU(inplace=inplace_relu)
        self.branch1 = nn.Conv3d(
            dim_in,
            dim_out,
            kernel_size=1,
            stride=[temporal_stride, spatial_stride, spatial_stride],
            padding=0,
            bias=False,
        )

    def forward(self, x):
        if hasattr(self, "branch1_bn") and hasattr(self, "branch1_relu"):
            x = self.branch1_relu(self.branch1_bn(x))
        x = self.branch1(x)
        return x


skip_transformations = {
    "postactivated_shortcut": PostactivatedShortcutTransformation,
    "preactivated_shortcut": PreactivatedShortcutTransformation,
    # For more types of skip transformations, add them below
}


class ResBlock(nn.Module):
    """
    Residual block with skip connection.
    """

    def __init__(
        self,
        dim_in,
        dim_out,
        dim_inner,
        temporal_kernel_size,
        temporal_conv_1x1,
        temporal_stride,
        spatial_stride,
        skip_transformation_type,
        residual_transformation_type,
        num_groups=1,
        inplace_relu=True,
        bn_eps=1e-5,
        bn_mmt=0.1,
        disable_pre_activation=False,
    ):
        """
        ResBlock class constructs redisual blocks. More details can be found in:
            "Deep residual learning for image recognition."
            https://arxiv.org/abs/1512.03385
        Args:
            dim_in (int): the channel dimensions of the input.
            dim_out (int): the channel dimension of the output.
            dim_inner (int): the inner dimension of the block.
            temporal_kernel_size (int): the temporal kernel sizes of the middle
                convolution in the bottleneck.
            temporal_conv_1x1 (bool): Only useful for PostactivatedBottleneckTransformation.
                if True, do temporal convolution in the fist 1x1 Conv3d.
                Otherwise, do it in the second 3x3 Conv3d
            temporal_stride (int): the temporal stride of the bottleneck.
            spatial_stride (int): the spatial_stride of the bottleneck.
            stride (int): the stride of the bottleneck.
            skip_transformation_type (str): the type of skip transformation
            residual_transformation_type (str): the type of residual transformation
            num_groups (int): number of groups for the convolution. num_groups=1
                is for standard ResNet like networks, and num_groups>1 is for
                ResNeXt like networks.
            disable_pre_activation (bool): If true, disable the preactivation,
                which includes BatchNorm3D and ReLU.
        """
        super(ResBlock, self).__init__()

        assert skip_transformation_type in skip_transformations, (
            "unknown skip transformation: %s" % skip_transformation_type
        )

        if (dim_in != dim_out) or (spatial_stride != 1) or (temporal_stride != 1):
            self.skip = skip_transformations[skip_transformation_type](
                dim_in,
                dim_out,
                temporal_stride,
                spatial_stride,
                bn_eps=bn_eps,
                bn_mmt=bn_mmt,
                disable_pre_activation=disable_pre_activation,
            )

        assert residual_transformation_type in residual_transformations, (
            "unknown residual transformation: %s" % residual_transformation_type
        )
        self.residual = residual_transformations[residual_transformation_type](
            dim_in,
            dim_out,
            temporal_stride,
            spatial_stride,
            num_groups,
            dim_inner,
            temporal_kernel_size=temporal_kernel_size,
            temporal_conv_1x1=temporal_conv_1x1,
            disable_pre_activation=disable_pre_activation,
        )
        self.relu = nn.ReLU(inplace_relu)

    def forward(self, x):
        if hasattr(self, "skip"):
            x = self.skip(x) + self.residual(x)
        else:
            x = x + self.residual(x)
        x = self.relu(x)
        return x



import torch
import torch.nn as nn
from .model import Model


class EncoderDecoder(Model):

    def __init__(self, encoder, decoder, activation):
        super().__init__()
        self.encoder = encoder
        self.decoder = decoder

        if callable(activation) or activation is None:
            self.activation = activation
        elif activation == 'softmax':
            self.activation = nn.Softmax(dim=1)
        elif activation == 'sigmoid':
            self.activation = nn.Sigmoid()
        else:
            raise ValueError(
                'Activation should be "sigmoid"/"softmax"/callable/None')

    def forward(self, x):
        """Sequentially pass `x` trough model`s
         `encoder` and `decoder` (return logits!)"""
        x = self.encoder(x)
        x = self.decoder(x)
        return x

    def predict(self, x):
        """Inference method. Switch model to `eval` mode, call `.forward(x)`
        and apply activation function
         (if activation is not `None`) with `torch.no_grad()`

        Args:
            x: 4D torch tensor with shape (batch_size, channels, height, width)

        Return:
            prediction: 4D torch tensor with shape
            (batch_size, classes, height, width)

        """
        if self.training:
            self.eval()

        with torch.no_grad():
            x = self.forward(x)
            if self.activation:
                x = self.activation(x)

        return x



from mlcomp.db.models import ReportTasks
from mlcomp.db.providers.base import BaseDataProvider


class ReportTasksProvider(BaseDataProvider):
    model = ReportTasks


__all__ = ['ReportTasksProvider']



from collections import defaultdict
from typing import List, Iterator
import numpy as np
from torch.utils.data import Sampler


class BalanceClassSampler(Sampler):
    """
    Abstraction over data sampler. Allows you to create stratified sample
    on unbalanced classes.
    """

    def __init__(
            self,
            labels: List,
            mode: str = 'downsampling',
            count_per_class: dict = None,
            max_count: int = None
    ):
        """
        Args:
            labels (List[int]): list of class label
                for each elem in the datasety
            mode (str): Strategy to balance classes.
                Must be one of [downsampling, upsampling]
        """
        samples_per_class = defaultdict(int)
        for l in labels:
            samples_per_class[l] += 1

        self.lbl2idx = dict()
        for i, l in enumerate(labels):
            if l not in self.lbl2idx:
                self.lbl2idx[l] = []
            self.lbl2idx[l].append(i)

        if mode == 'upsampling' or max_count is not None:
            samples_per_class = max_count \
                if max_count is not None \
                else max(samples_per_class.values())
        else:
            samples_per_class = min(samples_per_class.values())

        if count_per_class is not None:
            self.count_per_class = count_per_class
        else:
            self.count_per_class = {l: samples_per_class for l in self.lbl2idx}
        self.labels = labels
        self.length = sum(self.count_per_class.values())

        super().__init__(labels)

    def __iter__(self) -> Iterator[int]:
        """
        Yields:
            indices of stratified sample
        """
        indices = []
        for key in sorted(self.lbl2idx):
            samples = self.count_per_class[key]
            replace_ = samples > len(self.lbl2idx[key])
            indices += np.random.choice(
                self.lbl2idx[key], samples, replace=replace_
            ).tolist()
        assert (len(indices) == self.length)
        np.random.shuffle(indices)

        return iter(indices)

    def __len__(self) -> int:
        """
        Returns:
             length of result sample
        """
        return self.length


__all__ = ['BalanceClassSampler']



import os

import numpy as np

from collections import defaultdict
from catalyst.dl import State as RunnerState
from catalyst.dl.core import Callback


class InferBestCallback(Callback):
    def __init__(self, out_dir=None, out_prefix=None, best_only=False):
        self.out_dir = out_dir
        self.out_prefix = out_prefix
        self.predictions = defaultdict(lambda: [])
        self.best_only = best_only
        self._keys_from_state = ['out_dir', 'out_prefix']

    def on_stage_start(self, state: RunnerState):
        for key in self._keys_from_state:
            value = getattr(state, key, None)
            if value is not None:
                setattr(self, key, value)
        # assert self.out_prefix is not None
        if self.out_dir is not None:
            self.out_prefix = str(self.out_dir) + '/' + str(self.out_prefix)
        if self.out_prefix is not None:
            os.makedirs(os.path.dirname(self.out_prefix), exist_ok=True)

    def on_loader_start(self, state: RunnerState):
        self.predictions = defaultdict(lambda: [])

    def on_batch_end(self, state: RunnerState):
        dct = state.output
        dct = {key: value.detach().cpu().numpy() for key, value in dct.items()}
        for key, value in dct.items():
            self.predictions[key].append(value)

    def on_loader_end(self, state: RunnerState):
        if self.best_only and not state.metrics.is_best:
            return

        self.predictions = {
            key: np.concatenate(value, axis=0)
            for key, value in self.predictions.items()
        }
        if self.out_prefix is not None:
            for key, value in self.predictions.items():
                suffix = '.'.join([state.loader_name, key])
                np.save(f'{self.out_prefix}/{suffix}.npy', value)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter
from torch.nn.modules.loss import CrossEntropyLoss


class RingLoss(nn.Module):
    def __init__(self, type='auto', loss_weight=1.0, softmax_loss_weight=1.0):
        """
        :param type: type of loss ('l1', 'l2', 'auto')
        :param loss_weight: weight of loss, for 'l1' and 'l2', try with 0.01.
            For 'auto', try with 1.0.

        Source: https://github.com/Paralysis/ringloss
        """
        super().__init__()
        self.radius = Parameter(torch.Tensor(1))
        self.radius.data.fill_(-1)
        self.loss_weight = loss_weight
        self.type = type
        self.softmax = CrossEntropyLoss()
        self.softmax_loss_weight = softmax_loss_weight

    def forward(self, x, y):
        softmax = self.softmax(x, y).mul_(self.softmax_loss_weight)
        x = x.pow(2).sum(dim=1).pow(0.5)
        if self.radius.data[0] < 0:
            self.radius.data.fill_(x.mean().data)
        if self.type == 'l1':
            loss1 = F.smooth_l1_loss(x, self.radius.expand_as(x)). \
                mul_(self.loss_weight)
            loss2 = F.smooth_l1_loss(self.radius.expand_as(x), x). \
                mul_(self.loss_weight)
            ringloss = loss1 + loss2
        elif self.type == 'auto':
            diff = x.sub(self.radius.expand_as(x)) / \
                   (x.mean().detach().clamp(min=0.5))
            diff_sq = torch.pow(torch.abs(diff), 2).mean()
            ringloss = diff_sq.mul_(self.loss_weight)
        else:  # L2 Loss, if not specified
            diff = x.sub(self.radius.expand_as(x))
            diff_sq = torch.pow(torch.abs(diff), 2).mean()
            ringloss = diff_sq.mul_(self.loss_weight)
        return softmax + ringloss


__all__ = ['RingLoss']



import sqlalchemy as sa
from sqlalchemy import ForeignKey

from mlcomp.db.models.base import Base
from mlcomp.utils.misc import now


class Computer(Base):
    __tablename__ = 'computer'

    name = sa.Column(sa.String, primary_key=True)
    gpu = sa.Column(sa.Integer, default=0)
    cpu = sa.Column(sa.Integer, default=1)
    memory = sa.Column(sa.Float, default=0.1)
    usage = sa.Column(sa.String)
    ip = sa.Column(sa.String)
    port = sa.Column(sa.Integer)
    user = sa.Column(sa.String)
    last_synced = sa.Column(sa.DateTime)
    disk = sa.Column(sa.Integer)
    syncing_computer = sa.Column(sa.String, ForeignKey('computer.name'))
    root_folder = sa.Column(sa.String)
    can_process_tasks = sa.Column(sa.Boolean)
    sync_with_this_computer = sa.Column(sa.Boolean)
    meta = sa.Column(sa.String)


class ComputerUsage(Base):
    __tablename__ = 'computer_usage'

    id = sa.Column(sa.Integer, primary_key=True)
    computer = sa.Column(sa.String, ForeignKey('computer.name'))
    usage = sa.Column(sa.String)
    time = sa.Column(sa.DateTime, default=now())


__all__ = ['Computer', 'ComputerUsage']



import sys

import sqlalchemy as sa
from sqlalchemy import ForeignKey

from mlcomp.db.models.base import Base


class File(Base):
    __tablename__ = 'file'

    id = sa.Column(sa.Integer, primary_key=True)
    md5 = sa.Column(sa.String)
    created = sa.Column(sa.DateTime, default='Now()')
    content = sa.Column(sa.LargeBinary)
    project = sa.Column(sa.Integer, ForeignKey('project.id'))
    dag = sa.Column(sa.Integer, ForeignKey('dag.id'))
    size = sa.Column(sa.BigInteger, nullable=False, default=0)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size = sys.getsizeof(self.content)


__all__ = ['File']



from sqlalchemy import Table, Column, Integer, String, MetaData, Float, \
    TIMESTAMP, Boolean, LargeBinary, Index, BigInteger
from migrate.changeset.constraint import ForeignKeyConstraint, UniqueConstraint

meta = MetaData()

auxiliary = Table(
    'auxiliary', meta,
    Column('name', String(100), primary_key=True),
    Column('data', String(16000), nullable=False)
)

computer = Table(
    'computer', meta,
    Column('name', String(100), primary_key=True),
    Column('gpu', Integer, default=0, nullable=False),
    Column('cpu', Integer, default=1, nullable=False),
    Column('memory', Float, default=0.1, nullable=False),
    Column('usage', String(2000)),
    Column('ip', String(100), nullable=False),
    Column('port', Integer, nullable=False),
    Column('user', String, nullable=False),
    Column('last_synced', TIMESTAMP),
    Column('disk', Integer, nullable=False),
    Column('syncing_computer', String(100)),
    Column('root_folder', String(100), nullable=False),
    Column('can_process_tasks', Boolean),
    Column('sync_with_this_computer', Boolean),
    Column('meta', String(10000)),
)

computer_usage = Table(
    'computer_usage', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('computer', String(100), nullable=False),
    Column('usage', String(4000), nullable=False),
    Column('time', TIMESTAMP, nullable=False, default='now()')
)

dag = Table(
    'dag', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('name', String(100), nullable=False),
    Column('created', TIMESTAMP, nullable=False, default='now()'),
    Column('config', String(8000), nullable=False),
    Column('project', Integer, nullable=False),
    Column('docker_img', String(100)),
    Column('img_size', BigInteger, nullable=False),
    Column('file_size', BigInteger, nullable=False),
    Column('type', Integer, nullable=False),
    Column('report', Integer)
)

dag_library = Table(
    'dag_library', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('library', String(100), nullable=False),
    Column('version', String(30), nullable=False),
    Column('dag', Integer, nullable=False)
)

dag_storage = Table(
    'dag_storage', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('dag', Integer, nullable=False),
    Column('file', Integer),
    Column('path', String(210), nullable=False),
    Column('is_dir', Boolean, nullable=False, default=False)
)

file = Table(
    'file', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('md5', String(32), nullable=False),
    Column('created', TIMESTAMP, nullable=False, default='now()'),
    Column('content', LargeBinary, nullable=False),
    Column('project', Integer, nullable=False),
    Column('dag', Integer, nullable=False),
    Column('size', BigInteger, nullable=False),
)

log = Table(
    'log', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('step', Integer),
    Column('message', String(8000), nullable=False),
    Column('time', TIMESTAMP, nullable=False, default='now()'),
    Column('level', Integer, nullable=False),
    Column('component', Integer, nullable=False),
    Column('module', String(200), nullable=False),
    Column('line', Integer, nullable=False),
    Column('task', Integer),
    Column('computer', String(100))
)

project = Table(
    'project', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('name', String(180), nullable=False),
    Column('class_names', String(8000), nullable=False),
    Column('ignore_folders', String(8000), nullable=False)
)

report = Table(
    'report', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('config', String(4000), nullable=False),
    Column('time', TIMESTAMP, nullable=False, default='now()'),
    Column('name', String(300), nullable=False),
    Column('project', Integer, nullable=False),
    Column('layout', String(100), nullable=False)
)

report_img = Table(
    'report_img', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('epoch', Integer, nullable=False),
    Column('task', Integer, nullable=False),
    Column('group', String(100), nullable=False),
    Column('img', LargeBinary, nullable=False),
    Column('project', Integer, nullable=False),
    Column('dag', Integer, nullable=False),
    Column('part', String(30)),
    Column('y_pred', Integer),
    Column('y', Integer),
    Column('score', Float),
    Column('attr1', Float),
    Column('attr2', Float),
    Column('attr3', Float),
    Column('attr4', Float),
    Column('attr5', Float),
    Column('attr6', Float),
    Column('attr7', Float),
    Column('attr8', Float),
    Column('attr9', Float),
    Column('attr1_str', String(500)),
    Column('attr2_str', String(500)),
    Column('attr3_str', String(500)),
    Column('attr4_str', String(500)),
    Column('attr5_str', String(500)),
    Column('size', BigInteger, nullable=False),
)

report_layout = Table(
    'report_layout', meta,
    Column('name', String(400), primary_key=True),
    Column('content', String(8000), nullable=False),
    Column('last_modified', TIMESTAMP, nullable=False, default='now()')
)

report_series = Table(
    'report_series', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('task', Integer, nullable=False),
    Column('time', TIMESTAMP, nullable=False, default='now()'),
    Column('epoch', Integer, nullable=False),
    Column('value', Float, nullable=False),
    Column('name', String(100), nullable=False),
    Column('part', String(50), nullable=False),
    Column('stage', String(100), nullable=False)
)

report_task = Table(
    'report_task', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('report', Integer, nullable=False),
    Column('task', Integer, nullable=False),
)

step = Table(
    'step', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('task', Integer, nullable=False),
    Column('level', Integer, nullable=False),
    Column('started', TIMESTAMP, nullable=False),
    Column('finished', TIMESTAMP),
    Column('name', String(100), nullable=False),
    Column('index', Integer, nullable=False)
)

task = Table(
    'task', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('name', String(300), nullable=False),
    Column('status', Integer, nullable=False),
    Column('started', TIMESTAMP),
    Column('finished', TIMESTAMP),
    Column('computer', String(100)),
    Column('gpu', Integer, nullable=False, default=0),
    Column('gpu_max', Integer, nullable=False, default=0),
    Column('cpu', Integer, nullable=False, default=1),
    Column('executor', String(100), nullable=False),
    Column('computer_assigned', String(100)),
    Column('memory', Float, nullable=False, default=0.1),
    Column('steps', Integer, nullable=False, default=1),
    Column('current_step', String(100)),
    Column('dag', Integer, nullable=False),
    Column('celery_id', String(50)),
    Column('last_activity', TIMESTAMP),
    Column('debug', Boolean, nullable=False, default=False),
    Column('pid', Integer),
    Column('worker_index', Integer),
    Column('additional_info', String(16000), nullable=False),
    Column('docker_assigned', String(100)),
    Column('type', Integer, nullable=False),
    Column('score', Float),
    Column('report', Integer),
    Column('gpu_assigned', String(200)),
    Column('parent', Integer),
    Column('result', String(16000)),

    Column('batch_index', Integer),
    Column('batch_total', Integer),
    Column('loader_name', String),
    Column('epoch_duration', Integer),
    Column('epoch_time_remaining', Integer),
)

task_dependency = Table(
    'task_dependency', meta,
    Column('task_id', Integer, primary_key=True),
    Column('depend_id', Integer, primary_key=True),
)

task_synced = Table(
    'task_synced', meta,
    Column('computer', String(100), primary_key=True),
    Column('task', Integer, primary_key=True),
)

docker = Table(
    'docker', meta,
    Column('name', String(100), primary_key=True),
    Column('computer', String(100), primary_key=True),
    Column('last_activity', TIMESTAMP, nullable=False, default='now()'),
    Column('ports', String(100), nullable=False)
)

model = Table(
    'model', meta,
    Column('id', Integer, primary_key=True, autoincrement=True),
    Column('name', String(500), nullable=False),
    Column('score_local', Float),
    Column('score_public', Float),
    Column('dag', Integer),
    Column('project', Integer, nullable=False),
    Column('created', TIMESTAMP, nullable=False),
    Column('equations', String(2000000), nullable=False),
    Column('fold', Integer, nullable=False)
)


def upgrade(migrate_engine):
    conn = migrate_engine.connect()
    trans = conn.begin()

    try:
        meta.bind = conn

        computer.create()
        computer_usage.create()
        dag.create()
        dag_library.create()
        dag_storage.create()
        file.create()
        log.create()
        project.create()
        report.create()
        report_img.create()
        report_series.create()
        report_task.create()
        step.create()
        task.create()
        task_dependency.create()
        report_layout.create()
        docker.create()
        model.create()
        auxiliary.create()
        task_synced.create()

        ForeignKeyConstraint([computer_usage.c.computer],
                             [computer.c.name],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint([computer.c.syncing_computer],
                             [computer.c.name],
                             ondelete='CASCADE').create()

        Index('computer_usage_time_idx', computer_usage.c.time.desc()).create()

        ForeignKeyConstraint([dag.c.project], [project.c.id],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint([dag.c.report], [report.c.id],
                             ondelete='CASCADE').create()

        Index('dag_project_idx', dag.c.project.desc()).create()
        Index('dag_created_idx', dag.c.created.desc()).create()
        ForeignKeyConstraint([dag_library.c.dag], [dag.c.id],
                             ondelete='CASCADE').create()
        Index('dag_library_dag_idx', dag_library.c.dag.desc()).create()

        ForeignKeyConstraint([dag_storage.c.dag], [dag.c.id],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint([dag_storage.c.file], [file.c.id],
                             ondelete='CASCADE').create()
        Index('dag_storage_dag_idx', dag_storage.c.dag.desc()).create()

        ForeignKeyConstraint([file.c.project], [project.c.id],
                             ondelete='CASCADE').create()
        Index('file_created_idx', file.c.created.desc()).create()
        Index('file_project_idx', file.c.project.desc()).create()
        UniqueConstraint(file.c.md5, file.c.project,
                         name='file_md5_idx').create()

        ForeignKeyConstraint([log.c.task], [task.c.id],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint([log.c.step], [step.c.id],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint([log.c.computer], [computer.c.name],
                             ondelete='CASCADE').create()

        Index('log_step_idx', log.c.step.desc()).create()
        Index('log_time_idx', log.c.time.desc()).create()

        UniqueConstraint(project.c.name, name='project_name').create()

        ForeignKeyConstraint([report.c.project], [project.c.id],
                             ondelete='CASCADE').create()

        ForeignKeyConstraint([report_img.c.project], [project.c.id],
                             ondelete='CASCADE').create()
        Index('report_img_project_idx', report_img.c.project.desc()).create()
        Index('report_img_task_idx', report_img.c.task.desc()).create()
        Index('report_img_attr1_idx', report_img.c.attr1).create()
        Index('report_img_attr2_idx', report_img.c.attr2).create()
        Index('report_img_attr3_idx', report_img.c.attr3).create()
        Index('report_img_attr4_idx', report_img.c.attr4).create()
        Index('report_img_attr5_idx', report_img.c.attr5).create()
        Index('report_img_attr6_idx', report_img.c.attr6).create()
        Index('report_img_attr7_idx', report_img.c.attr7).create()
        Index('report_img_attr8_idx', report_img.c.attr8).create()
        Index('report_img_attr9_idx', report_img.c.attr9).create()
        Index('report_img_attr1_str_idx', report_img.c.attr1_str).create()
        Index('report_img_attr2_str_idx', report_img.c.attr2_str).create()
        Index('report_img_attr3_str_idx', report_img.c.attr3_str).create()
        Index('report_img_attr4_str_idx', report_img.c.attr4_str).create()
        Index('report_img_attr5_str_idx', report_img.c.attr5_str).create()

        ForeignKeyConstraint([report_series.c.task], [task.c.id],
                             ondelete='CASCADE').create()

        ForeignKeyConstraint([report_task.c.task], [task.c.id],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint([report_task.c.report], [report.c.id],
                             ondelete='CASCADE').create()

        ForeignKeyConstraint([step.c.task], [task.c.id],
                             ondelete='CASCADE').create()
        Index('step_name_idx', step.c.name).create()

        ForeignKeyConstraint([task.c.computer], [computer.c.name],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint([task.c.report], [report.c.id],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint([task.c.computer_assigned], [computer.c.name],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint([task.c.dag], [dag.c.id],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint([task.c.parent], [task.c.id],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint(
            [task.c.docker_assigned, task.c.computer_assigned],
            [docker.c.name, docker.c.computer], ondelete='CASCADE').create()

        Index('log_message_idx', log.c.message).create()
        Index('task_status_idx', task.c.status).create()
        Index('task_dag_idx', task.c.dag.desc()).create()
        Index('task_finished_idx', task.c.finished.desc()).create()
        Index('task_name_idx', task.c.name).create()
        Index('task_started_idx', task.c.started.desc()).create()
        Index('task_docker_idx', task.c.docker_assigned,
              task.c.computer_assigned).create()

        Index('task_dependency_task_idx',
              task_dependency.c.task_id.desc()).create()
        Index('task_dependency_depend_idx',
              task_dependency.c.depend_id.desc()).create()

        ForeignKeyConstraint([task_dependency.c.task_id], [task.c.id],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint([task_dependency.c.depend_id], [task.c.id],
                             ondelete='CASCADE').create()

        ForeignKeyConstraint([docker.c.computer], [computer.c.name],
                             ondelete='CASCADE').create()

        ForeignKeyConstraint([model.c.dag], [dag.c.id],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint([model.c.project], [project.c.id],
                             ondelete='CASCADE').create()
        UniqueConstraint(model.c.project, model.c.name,
                         name='model_project_name_unique').create()

        ForeignKeyConstraint([task_synced.c.computer], [computer.c.name],
                             ondelete='CASCADE').create()
        ForeignKeyConstraint([task_synced.c.task], [task.c.id],
                             ondelete='CASCADE').create()
    except Exception:
        trans.rollback()
        raise
    else:
        trans.commit()


def downgrade(migrate_engine):
    conn = migrate_engine.connect()
    trans = conn.begin()
    meta.bind = conn
    try:
        auxiliary.drop()
        model.drop()
        docker.drop()
        computer_usage.drop()
        dag_library.drop()
        dag_storage.drop()
        log.drop()
        file.drop()
        report_task.drop()
        report_img.drop()
        report_series.drop()
        report_layout.drop()
        report.drop()
        task_dependency.drop()
        step.drop()
        task.drop()
        dag.drop()
        computer.drop()
        project.drop()
    except Exception:
        trans.rollback()
        raise
    else:
        trans.commit()



from mlcomp.db.core import PaginatorOptions
from mlcomp.db.models import ReportLayout
from mlcomp.db.providers.base import BaseDataProvider
from mlcomp.db.report_info import ReportLayoutInfo
from mlcomp.utils.misc import now
from mlcomp.utils.io import yaml_load, yaml_dump


class ReportLayoutProvider(BaseDataProvider):
    model = ReportLayout

    def get(self, filter: dict = None, options: PaginatorOptions = None):
        query = self.query(ReportLayout)
        total = query.count()
        paginator = self.paginator(query, options)

        res = []
        for item in paginator.all():
            res.append(self.to_dict(item))

        return {'total': total, 'data': res}

    def by_name(self, name: str):
        return self.query(ReportLayout).filter(ReportLayout.name == name).one()

    def add_item(self, k: str, v: dict):
        self.add(
            ReportLayout(content=yaml_dump(v), name=k, last_modified=now())
        )

    def all(self):
        res = {
            s.name: yaml_load(s.content)
            for s in self.query(ReportLayout).all()
        }

        for k, v in res.items():
            res[k] = ReportLayoutInfo.union_layouts(k, res)
        return res

    def change(self, k: str, v: dict):
        self.query(ReportLayout).filter(ReportLayout.name == k).update(
            {
                'last_modified': now(),
                'content': yaml_dump(v)
            }
        )


__all__ = ['ReportLayoutProvider']



import pickle
from copy import deepcopy

import numpy as np
import pandas as pd

from mlcomp.worker.executors import Executor

from mlcomp.worker.executors.infer import Infer
from mlcomp.worker.reports.classification import ClassificationReportBuilder

from dataset import MnistDataset
from experiment import Experiment


@Executor.register
class InferMnist(Infer):
    def __init__(self, **kwargs):
        cache_names = ['y']
        super().__init__(cache_names=cache_names, layout='img_classify',
                         **kwargs)

        if self.test:
            self.x_source = MnistDataset(
                file='data/test.csv',
                transforms=Experiment.get_test_transforms(),
                max_count=self.max_count,
            )
        else:
            self.x_source = MnistDataset(
                file='data/train.csv',
                fold_csv='data/fold.csv',
                is_test=True,
                transforms=Experiment.get_test_transforms(),
                max_count=self.max_count
            )

        self.builder = None
        self.x = None
        self.res = []
        self.submit_res = []

    def create_base(self):
        self.builder = ClassificationReportBuilder(
            session=self.session,
            task=self.task,
            layout=self.layout,
            name=self.name,
            plot_count=self.plot_count
        )
        self.builder.create_base()

    def count(self):
        return len(self.x_source)

    def adjust_part(self, part):
        self.x = deepcopy(self.x_source)
        self.x.x = self.x.x[part[0]:part[1]]
        if not self.test:
            self.x.y = self.x.y[part[0]:part[1]]

    def save(self, preds, folder: str):
        self.res.extend(preds)

    def save_final(self, folder):
        pickle.dump(np.array(self.res),
                    open(f'{folder}/{self.model_name}_{self.suffix}.p', 'wb'))

    def submit(self, preds):
        argmax = preds.argmax(axis=1)
        self.submit_res.extend(
            [{'ImageId': len(self.submit_res) + i + 1, 'Label': p} for i, p in
             enumerate(argmax)])

    def submit_final(self, folder):
        pd.DataFrame(self.submit_res). \
            to_csv(f'{folder}/{self.model_name}_{self.suffix}.csv',
                   index=False)

    def _plot_main(self, preds):
        imgs = [
            ((row['features'][0] * 0.229 + 0.485) * 255).astype(np.uint8)
            for row in self.x
        ]
        attrs = [
            {
                'attr1': p.argmax()
            } for p in preds
        ]

        self.builder.process_pred(
            imgs=imgs,
            preds=preds,
            attrs=attrs
        )

    def plot(self, preds):
        self._plot_main(preds)



from itertools import product
from os.path import join
from typing import List
from glob import glob

from mlcomp.utils.io import yaml_load
from mlcomp.utils.misc import dict_flatten


def cell_name(cell: dict):
    c = dict_flatten(cell)
    parts = []
    for k, v in c.items():
        parts.append(f'{k}={v}')

    return ' '.join(parts)[-300:]


def grid_cells(grid: List):
    for i, row in enumerate(grid):
        row_type = type(row)

        if row_type == list:
            if len(row) == 0:
                raise Exception(f'Empty list at {i} position')
            if type(row[0]) != dict:
                raise Exception('List entries can be dicts only')
        elif row_type == dict:
            if len(row) != 1:
                raise Exception('Dict must contain only one element')
            key = list(row)[0]
            val_type = type(row[key])
            if val_type not in [list, str]:
                raise Exception('Dict value must be list or str')
            new_row = []
            if val_type == str:
                if '-' in row[key]:
                    start, end = map(int, row[key].split('-'))
                    for p in range(start, end + 1):
                        new_row.append({key: p})
                else:
                    if key == '_folder':
                        for file in glob(join(row[key], '*.yml')):
                            new_row.append(yaml_load(file))
            else:
                for v in row[key]:
                    if key == '_file':
                        new_row.append(yaml_load(v))
                    else:
                        new_row.append({key: v})

            grid[i] = new_row
        else:
            raise Exception(f'Unknown type of row = {row_type}')

    res = list(product(*grid))
    for i, r in enumerate(res):
        d = {}
        for dd in r:
            d.update(dd)
        res[i] = d
    return [[r, cell_name(r)] for r in res]


__all__ = ['grid_cells']



from typing import List

from sqlalchemy.orm.query import Query
from sqlalchemy import desc
from sqlalchemy_serializer import Serializer
from sqlalchemy.orm import joinedload

from mlcomp.db.core import Session, PaginatorOptions
from mlcomp.db.models.base import Base
from mlcomp.utils.misc import adapt_db_types


class BaseDataProvider:
    model = None

    date_format = '%Y-%m-%d'
    datetime_format = '%Y-%m-%d %H:%M:%SZ'
    time_format = '%H:%M'

    def __init__(self, session: Session = None):
        if session is None:
            session = Session.create_session()
        self._session = session
        self.serializer = Serializer(
            date_format=self.date_format,
            datetime_format=self.datetime_format,
            time_format=self.time_format
        )

    def serialize_datetime(self, value):
        return self.serializer.serialize_datetime(value)

    def remove(self, key_value, key_column: str = 'id'):
        self.query(self.model). \
            filter(getattr(self.model, key_column) == key_value). \
            delete(synchronize_session=False)
        self.session.commit()

    def detach(self, obj):
        self.session.expunge(obj)

    @property
    def query(self):
        return self.session.query

    def add_all(self, obs: List[Base], commit=True):
        self._session.add_all(obs, commit=commit)

    def add(self, obj: Base, commit=True):
        self._session.add(obj, commit=commit)
        return obj

    def bulk_save_objects(
        self,
        objs,
        return_defaults=False,
        update_changed_only=True,
        preserve_order=True,
    ):
        for obj in objs:
            adapt_db_types(obj)

        self._session.bulk_save_objects(
            objs,
            return_defaults=return_defaults,
            update_changed_only=update_changed_only,
            preserve_order=preserve_order
        )
        self._session.commit()

    def by_id(self, id: int, joined_load=None):
        res = self.query(self.model).filter(getattr(self.model, 'id') == id)
        if joined_load is not None:
            for n in joined_load:
                res = res.options(joinedload(n, innerjoin=True))
        return res.first()

    def to_dict(self, item, rules=(), datetime_format=None):
        datetime_format = datetime_format or self.datetime_format
        return item.to_dict(
            date_format=self.date_format,
            datetime_format=datetime_format,
            time_format=self.time_format,
            rules=rules
        )

    def create_or_update(self, obj: Base, *fields):
        query = self.session.query(obj.__class__)
        for k in fields:
            query = query.filter(getattr(obj.__class__, k) == getattr(obj, k))

        db = query.first()
        if db is not None:
            for field, value in obj.__dict__.items():
                if not field.startswith('_'):
                    setattr(db, field, value)
            self.session.update()
        else:
            self.add(obj)

    def all(self):
        return self.query(self.model).all()

    def update(self):
        self.session.update()
        self.session.commit()

    def commit(self):
        self.session.commit()

    def rollback(self):
        self.session.rollback()

    @property
    def session(self):
        return self._session

    def paginator(self, query: Query, options: PaginatorOptions):
        if options is None:
            return query

        if options.sort_column:
            column = getattr(self.model, options.sort_column) if \
                options.sort_column in self.model.__dict__ \
                else options.sort_column
            criterion = column if not options.sort_descending else desc(column)
            query = query.order_by(criterion)

        if options.page_size:
            query = query. \
                offset(options.page_size * options.page_number). \
                limit(options.page_size)

        return query


__all__ = ['BaseDataProvider']



import atexit

from apscheduler.schedulers.background import BackgroundScheduler


def start_schedule(jobs):
    scheduler = BackgroundScheduler()
    for func, interval in jobs:
        scheduler.add_job(func=func, trigger='interval', seconds=interval,
                          max_instances=1)
    scheduler.start()

    # Shut down the scheduler when exiting the app
    atexit.register(lambda: scheduler.shutdown())



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from collections import OrderedDict

import torch.nn as nn

from .resnext3d_block import ResBlock


class ResStageBase(nn.Module):
    def __init__(
        self,
        stage_idx,
        dim_in,
        dim_out,
        dim_inner,
        temporal_kernel_basis,
        temporal_conv_1x1,
        temporal_stride,
        spatial_stride,
        num_blocks,
        num_groups,
    ):
        super(ResStageBase, self).__init__()

        assert (
            len(
                {
                    len(dim_in),
                    len(dim_out),
                    len(temporal_kernel_basis),
                    len(temporal_conv_1x1),
                    len(temporal_stride),
                    len(spatial_stride),
                    len(num_blocks),
                    len(dim_inner),
                    len(num_groups),
                }
            )
            == 1
        )

        self.stage_idx = stage_idx
        self.num_blocks = num_blocks
        self.num_pathways = len(self.num_blocks)

        self.temporal_kernel_sizes = [
            (temporal_kernel_basis[i] * num_blocks[i])[: num_blocks[i]]
            for i in range(len(temporal_kernel_basis))
        ]

    def _block_name(self, pathway_idx, stage_idx, block_idx):
        return "pathway{}-stage{}-block{}".format(pathway_idx, stage_idx, block_idx)

    def _pathway_name(self, pathway_idx):
        return "pathway{}".format(pathway_idx)

    def forward(self, inputs):
        output = []
        for p in range(self.num_pathways):
            x = inputs[p]
            pathway_module = getattr(self, self._pathway_name(p))
            output.append(pathway_module(x))
        return output


class ResStage(ResStageBase):
    """
    Stage of 3D ResNet. It expects to have one or more tensors as input for
        single pathway (C2D, I3D, SlowOnly), and multi-pathway (SlowFast) cases.
        More details can be found here:
        "Slowfast networks for video recognition."
        https://arxiv.org/pdf/1812.03982.pdf
    """

    def __init__(
        self,
        stage_idx,
        dim_in,
        dim_out,
        dim_inner,
        temporal_kernel_basis,
        temporal_conv_1x1,
        temporal_stride,
        spatial_stride,
        num_blocks,
        num_groups,
        skip_transformation_type,
        residual_transformation_type,
        block_callback=None,
        inplace_relu=True,
        bn_eps=1e-5,
        bn_mmt=0.1,
        disable_pre_activation=False,
        final_stage=False,
    ):
        """
        The `__init__` method of any subclass should also contain these arguments.
        ResStage builds p streams, where p can be greater or equal to one.
        Args:
            stage_idx (int): integer index of stage.
            dim_in (list): list of p the channel dimensions of the input.
                Different channel dimensions control the input dimension of
                different pathways.
            dim_out (list): list of p the channel dimensions of the output.
                Different channel dimensions control the input dimension of
                different pathways.
            dim_inner (list): list of the p inner channel dimensions of the
                input.
                Different channel dimensions control the input dimension of
                different pathways.
            temporal_kernel_basis (list): Basis of temporal kernel sizes for each of
                the stage.
            temporal_conv_1x1 (list): Only useful for BottleneckBlock.
                In a pathaway, if True, do temporal convolution in the fist 1x1 Conv3d.
                Otherwise, do it in the second 3x3 Conv3d
            temporal_stride (list): the temporal stride of the bottleneck.
            spatial_stride (list): the spatial_stride of the bottleneck.
            num_blocks (list): list of p numbers of blocks for each of the
                pathway.
            num_groups (list): list of number of p groups for the convolution.
                num_groups=1 is for standard ResNet like networks, and
                num_groups>1 is for ResNeXt like networks.
            skip_transformation_type (str): the type of skip transformation
            residual_transformation_type (str): the type of residual transformation
            block_callback (function object): a callback function to be called with
                residual block and its name as input arguments
            disable_pre_activation (bool): If true, disable the preactivation,
                which includes BatchNorm3D and ReLU.
            final_stage (bool): If true, this is the last stage in the model.
        """
        super(ResStage, self).__init__(
            stage_idx,
            dim_in,
            dim_out,
            dim_inner,
            temporal_kernel_basis,
            temporal_conv_1x1,
            temporal_stride,
            spatial_stride,
            num_blocks,
            num_groups,
        )

        for p in range(self.num_pathways):
            blocks = []
            for i in range(self.num_blocks[p]):
                # Retrieve the transformation function.
                # Construct the block.
                block_disable_pre_activation = (
                    True if disable_pre_activation and i == 0 else False
                )
                res_block = ResBlock(
                    dim_in[p] if i == 0 else dim_out[p],
                    dim_out[p],
                    dim_inner[p],
                    self.temporal_kernel_sizes[p][i],
                    temporal_conv_1x1[p],
                    temporal_stride[p] if i == 0 else 1,
                    spatial_stride[p] if i == 0 else 1,
                    skip_transformation_type,
                    residual_transformation_type,
                    num_groups=num_groups[p],
                    inplace_relu=inplace_relu,
                    bn_eps=bn_eps,
                    bn_mmt=bn_mmt,
                    disable_pre_activation=block_disable_pre_activation,
                )
                block_name = self._block_name(p, stage_idx, i)
                if block_callback:
                    res_block = block_callback(block_name, res_block)
                blocks.append((block_name, res_block))

            if final_stage and (
                residual_transformation_type == "preactivated_bottleneck_transformation"
            ):
                # For pre-activation residual transformation, we conduct
                # activation in the final stage before continuing forward pass
                # through the head
                activate_bn = nn.BatchNorm3d(dim_out[p])
                activate_relu = nn.ReLU(inplace=True)
                activate_bn_name = "-".join([block_name, "bn"])
                activate_relu_name = "-".join([block_name, "relu"])
                if block_callback:
                    activate_relu = block_callback(activate_relu_name, activate_relu)
                blocks.append((activate_bn_name, activate_bn))
                blocks.append((activate_relu_name, activate_relu))

            self.add_module(self._pathway_name(p), nn.Sequential(OrderedDict(blocks)))



from mlcomp.contrib.segmentation.deeplabv3.backbone import resnet, \
    xception, drn, mobilenet


def build_backbone(backbone, output_stride, BatchNorm):
    if backbone == 'resnet':
        return resnet.ResNet101(output_stride, BatchNorm)
    elif backbone == 'xception':
        return xception.AlignedXception(output_stride, BatchNorm)
    elif backbone == 'drn':
        return drn.drn_d_54(BatchNorm)
    elif backbone == 'mobilenet':
        return mobilenet.MobileNetV2(output_stride, BatchNorm)
    else:
        raise NotImplementedError



class PaginatorOptions:
    def __init__(
        self,
        page_number: int,
        page_size: int,
        sort_column: str = None,
        sort_descending: bool = None
    ):
        self.sort_column = sort_column
        self.sort_descending = sort_descending
        self.page_number = page_number
        self.page_size = page_size

        assert (page_number is not None and page_size) \
            or (page_number is not None and not page_size), \
            'Specify both page_number and page_size'

        if not sort_column:
            self.sort_column = 'id'
            self.sort_descending = True


__all__ = ['PaginatorOptions']



import os
import shutil
import socket
import time
import traceback
import sys
from os.path import join, dirname, abspath
from typing import List

import psutil
from sqlalchemy.orm import joinedload
from celery.signals import celeryd_after_setup
from celery import states

from mlcomp import MODEL_FOLDER, TASK_FOLDER, DOCKER_IMG
from mlcomp.db.core import Session
from mlcomp.db.enums import ComponentType, TaskStatus
from mlcomp.db.models import Task, Dag
from mlcomp.db.providers import TaskProvider, \
    DagLibraryProvider, \
    DockerProvider
from mlcomp.utils.logging import create_logger
from mlcomp.utils.io import yaml_load, yaml_dump
from mlcomp.utils.misc import set_global_seed, now
from mlcomp.worker.app import app
from mlcomp.worker.executors import Executor
from mlcomp.worker.storage import Storage
from mlcomp.utils.config import Config


class ExecuteBuilder:
    def __init__(self, id: int, repeat_count: int = 1, exit=True):
        self.session = Session.create_session(key='ExecuteBuilder')
        self.id = id
        self.repeat_count = repeat_count
        self.logger = create_logger(self.session, 'ExecuteBuilder')
        self.logger_db = create_logger(self.session, 'ExecuteBuilder.db',
                                       console=False)
        self.exit = exit

        self.provider = None
        self.library_provider = None
        self.storage = None
        self.task = None
        self.dag = None
        self.executor = None
        self.hostname = None
        self.docker_img = None
        self.worker_index = None
        self.queue_personal = None
        self.config = None
        self.executor_type = None

    def info(self, msg: str, step=None):
        self.logger.info(msg, ComponentType.Worker, self.hostname, self.id,
                         step)

    def error(self, msg: str, step=None):
        self.logger.error(msg, ComponentType.Worker, self.hostname, self.id,
                          step)

    def warning(self, msg: str, step=None):
        self.logger.warning(msg, ComponentType.Worker, self.hostname, self.id,
                            step)

    def debug(self, msg: str, step=None):
        self.logger.debug(msg, ComponentType.Worker, self.hostname, self.id,
                          step)

    def create_base(self):
        self.info('create_base')

        if app.current_task:
            app.current_task.update_state(state=states.SUCCESS)
            app.control.revoke(app.current_task.request.id, terminate=True)

        self.provider = TaskProvider(self.session)
        self.library_provider = DagLibraryProvider(self.session)
        self.storage = Storage(self.session)

        self.task = self.provider.by_id(
            self.id, joinedload(Task.dag_rel, innerjoin=True)
        )
        if not self.task:
            raise Exception(f'task with id = {self.id} is not found')

        self.dag = self.task.dag_rel
        self.executor = None
        self.hostname = socket.gethostname()

        self.docker_img = DOCKER_IMG
        self.worker_index = os.getenv('WORKER_INDEX', -1)

        self.queue_personal = f'{self.hostname}_{self.docker_img}_' \
                              f'{self.worker_index}'

        self.config = Config.from_yaml(self.dag.config)

        set_global_seed(self.config['info'].get('seed', 0))

        self.executor_type = self.config['executors'][self.task.executor][
            'type']

        executor = self.config['executors'][self.task.executor]

        cuda_visible_devices = os.getenv('CUDA_VISIBLE_DEVICES', '')
        self.info(f'Env.before execution '
                  f'CUDA_VISIBLE_DEVICES={cuda_visible_devices}')

        if cuda_visible_devices.strip() != '':
            gpu_assigned = self.task.gpu_assigned or ''

            cuda_visible_devices = cuda_visible_devices.split(',')
            cuda_visible_devices = ','.join(
                [cuda_visible_devices[int(g)] for g in
                 gpu_assigned.split(',') if g.strip() != ''])
        else:
            cuda_visible_devices = self.task.gpu_assigned

        cuda_visible_devices = cuda_visible_devices or ''

        env = {
            'MKL_NUM_THREADS': 1,
            'OMP_NUM_THREADS': 1,
            'CUDA_VISIBLE_DEVICES': cuda_visible_devices
        }
        env.update(executor.get('env', {}))

        for k, v in env.items():
            os.environ[k] = str(v)
            self.info(f'Set env. {k} = {v}')

    def check_status(self):
        self.info('check_status')

        assert self.dag is not None, 'You must fetch task with dag_rel'

        if self.task.status >= TaskStatus.InProgress.value:
            msg = f'Task = {self.task.id}. Status = {self.task.status}, ' \
                  f'before the execute_by_id invocation.'
            if app.current_task:
                msg += f' Request Id = {app.current_task.request.id}'
            self.error(msg)
            return True

    def change_status(self):
        self.info('change_status')

        self.task.computer_assigned = self.hostname
        self.task.pid = os.getpid()
        self.task.worker_index = self.worker_index
        self.task.docker_assigned = self.docker_img
        self.task.status = TaskStatus.InProgress.value
        self.task.started = now()
        self.provider.commit()

    def download(self):
        self.info('download')

        if not self.task.debug:
            folder = self.storage.download(task=self.id)
        else:
            folder = os.getcwd()

        os.chdir(folder)

        libraries = self.library_provider.dag(self.task.dag)
        executor_type = self.executor_type

        self.info('download. folder changed')

        mlcomp_executors_folder = join(dirname(abspath(__file__)), 'executors')
        mlcomp_base_folder = os.path.abspath(join(mlcomp_executors_folder,
                                                  '../../../'))

        imported, was_installation = self.storage.import_executor(
            mlcomp_executors_folder,
            mlcomp_base_folder,
            executor_type)

        if not imported:
            imported, was_installation = self.storage.import_executor(
                folder,
                folder,
                executor_type,
                libraries)

            if not imported:
                raise Exception(f'Executor = {executor_type} not found')

        self.info('download. executor imported')

        if was_installation and not self.task.debug:
            if self.repeat_count > 0:
                self.info('was installation. '
                          'set task status to Queued. '
                          'And resending the task to a queue')
                self.task.status = TaskStatus.Queued.value
                self.provider.commit()

                try:
                    execute.apply_async(
                        (self.id, self.repeat_count - 1),
                        queue=self.queue_personal,
                        retry=False
                    )
                except Exception:
                    pass
                finally:
                    sys.exit()

        assert Executor.is_registered(executor_type), \
            f'Executor {executor_type} was not found'

    def create_executor(self):
        self.info('create_executor')

        additional_info = yaml_load(self.task.additional_info) \
            if self.task.additional_info else dict()
        self.executor = Executor.from_config(
            executor=self.task.executor, config=self.config,
            additional_info=additional_info,
            session=self.session, logger=self.logger, logger_db=self.logger_db
        )

    def execute(self):
        self.info('execute start')

        res = self.executor(task=self.task, task_provider=self.provider,
                            dag=self.dag)
        self.info('execute executor finished')

        res = res or {}
        self.task.result = yaml_dump(res)
        self.provider.commit()

        if 'stage' in res and 'stages' in res:
            index = res['stages'].index(res['stage'])
            if index < len(res['stages']) - 1:
                self.executor.info(
                    f'stage = {res["stage"]} done. '
                    f'Go to the stage = '
                    f'{res["stages"][index + 1]}'
                )

                time.sleep(3)

                self.executor.info(f'sending {(self.id, self.repeat_count)} '
                                   f'to {self.queue_personal}')

                self.task.status = TaskStatus.Queued.value
                self.provider.commit()

                execute.apply_async(
                    (self.id, self.repeat_count), queue=self.queue_personal,
                    retry=False
                )
                return

        self.executor.step.finish()
        self.provider.change_status(self.task, TaskStatus.Success)

        self.info('execute end')

    def build(self):
        try:
            self.create_base()

            bad_status = self.check_status()
            if bad_status:
                return

            self.change_status()

            self.download()

            self.create_executor()

            self.execute()

        except Exception as e:
            step = self.executor.step.id if \
                (self.executor and self.executor.step) else None

            if Session.sqlalchemy_error(e):
                Session.cleanup(key='ExecuteBuilder')
                self.session = Session.create_session(key='ExecuteBuilder')
                self.logger.session = create_logger(self.session,
                                                    'ExecuteBuilder')

            self.error(traceback.format_exc(), step)
            if self.task.status <= TaskStatus.InProgress.value:
                self.provider.change_status(self.task, TaskStatus.Failed)
            raise e
        finally:
            if app.current_task:
                app.close()

            if self.exit:
                # noinspection PyProtectedMember
                os._exit(0)


def execute_by_id(id: int, repeat_count=1, exit=True):
    ex = ExecuteBuilder(id, repeat_count=repeat_count, exit=exit)
    ex.build()


@celeryd_after_setup.connect
def capture_worker_name(sender, instance, **kwargs):
    os.environ['WORKER_INDEX'] = sender.split('_')[-1]


@app.task
def execute(id: int, repeat_count: int = 1):
    execute_by_id(id, repeat_count)


@app.task
def kill(pid: int):
    os.system(f'kill -9 {pid}')
    print('KILLED', pid)
    return not psutil.pid_exists(pid)


@app.task
def kill_all(pids: List[int]):
    os.system(f'kill -9 {" ".join(map(str, pids))}')
    return all(not psutil.pid_exists(pid) for pid in pids)


@app.task
def remove(path: str):
    if not os.path.exists(path):
        return
    if os.path.isdir(path):
        shutil.rmtree(path)
    else:
        os.remove(path)


def remove_from_all(session: Session, path: str):
    provider = DockerProvider(session)
    dockers = provider.get_online()
    for docker in dockers:
        queue = f'{docker.computer}_{docker.name or "default"}_supervisor'
        remove.apply((path,), queue=queue)


def remove_model(session: Session, project_name: str, model_name: str):
    path = join(MODEL_FOLDER, project_name, model_name + '.pth')
    path_weight = join(MODEL_FOLDER, project_name, model_name + '_weight.pth')

    remove_from_all(session, path)
    remove_from_all(session, path_weight)


def remove_task(session: Session, id: int):
    path = join(TASK_FOLDER, str(id))
    remove_from_all(session, path)


def remove_dag(session: Session, id: int):
    tasks = TaskProvider(session).by_dag(id)
    for task in tasks:
        remove_task(session, task.id)


def stop(logger, session: Session, task: Task, dag: Dag):
    provider = TaskProvider(session)
    if task.status > TaskStatus.InProgress.value:
        return task.status

    status = TaskStatus.Stopped
    try:
        if task.status != TaskStatus.NotRan.value:
            app.control.revoke(task.celery_id, terminate=True)
        else:
            status = TaskStatus.Skipped
    except Exception as e:
        if Session.sqlalchemy_error(e):
            try:
                logger.error(traceback.format_exc(), ComponentType.API)
            except Exception:
                pass
            raise
        logger.error(traceback.format_exc(), ComponentType.API)
    finally:
        if task.pid:
            queue = f'{task.computer_assigned}_' \
                    f'{dag.docker_img or "default"}_supervisor'
            kill.apply_async((task.pid,), queue=queue, retry=False)

            additional_info = yaml_load(task.additional_info)
            for p in additional_info.get('child_processes', []):
                kill.apply_async((p,), queue=queue, retry=False)
        provider.change_status(task, status)

    return task.status


if __name__ == '__main__':
    execute(81)
    # from task.tasks import execute
    # execute.delay(42)



import torch
import torch.nn as nn
import torch.nn.functional as F

from mlcomp.contrib.segmentation.deeplabv3.aspp import build_aspp
from mlcomp.contrib.segmentation.deeplabv3.backbone import build_backbone
from mlcomp.contrib.segmentation.deeplabv3.decoder import build_decoder


class DeepLab(nn.Module):
    def __init__(self, backbone='resnet', output_stride=16, num_classes=21,
                 freeze_bn=False):
        super(DeepLab, self).__init__()
        if backbone == 'drn':
            output_stride = 8

        BatchNorm = nn.BatchNorm2d
        self.backbone = build_backbone(backbone, output_stride, BatchNorm)
        self.aspp = build_aspp(backbone, output_stride, BatchNorm)
        self.decoder = build_decoder(num_classes, backbone, BatchNorm)

        if freeze_bn:
            self.freeze_bn()

    def forward(self, input):
        x, low_level_feat = self.backbone(input)
        x = self.aspp(x)
        x = self.decoder(x, low_level_feat)
        x = F.interpolate(x, size=input.size()[2:], mode='bilinear',
                          align_corners=True)

        return x

    def freeze_bn(self):
        for m in self.modules():
            if isinstance(m, nn.BatchNorm2d):
                m.eval()

    def get_1x_lr_params(self):
        modules = [self.backbone]
        for i in range(len(modules)):
            for m in modules[i].named_modules():
                if isinstance(m[1], nn.Conv2d) or \
                        isinstance(m[1], nn.BatchNorm2d):
                    for p in m[1].parameters():
                        if p.requires_grad:
                            yield p

    def get_10x_lr_params(self):
        modules = [self.aspp, self.decoder]
        for i in range(len(modules)):
            for m in modules[i].named_modules():
                if isinstance(m[1], nn.Conv2d) \
                        or isinstance(m[1], nn.BatchNorm2d):
                    for p in m[1].parameters():
                        if p.requires_grad:
                            yield p


if __name__ == '__main__':
    model = DeepLab(backbone='drn', output_stride=16)
    model.eval()
    input = torch.rand(1, 3, 513, 513)
    output = model(input)
    print(output.size())



import torch.nn as nn
from pretrainedmodels.models.inceptionresnetv2 import InceptionResNetV2
from pretrainedmodels.models.inceptionresnetv2 import pretrained_settings


class InceptionResNetV2Encoder(InceptionResNetV2):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.pretrained = False

        # correct paddings
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.kernel_size == (3, 3):
                    m.padding = (1, 1)
            if isinstance(m, nn.MaxPool2d):
                m.padding = (1, 1)

        # remove linear layers
        del self.avgpool_1a
        del self.last_linear

    def forward(self, x):
        x = self.conv2d_1a(x)
        x = self.conv2d_2a(x)
        x = self.conv2d_2b(x)
        x0 = x

        x = self.maxpool_3a(x)
        x = self.conv2d_3b(x)
        x = self.conv2d_4a(x)
        x1 = x

        x = self.maxpool_5a(x)
        x = self.mixed_5b(x)
        x = self.repeat(x)
        x2 = x

        x = self.mixed_6a(x)
        x = self.repeat_1(x)
        x3 = x

        x = self.mixed_7a(x)
        x = self.repeat_2(x)
        x = self.block8(x)
        x = self.conv2d_7b(x)
        x4 = x

        features = [x4, x3, x2, x1, x0]
        return features

    def load_state_dict(self, state_dict, **kwargs):
        state_dict.pop('last_linear.bias')
        state_dict.pop('last_linear.weight')
        super().load_state_dict(state_dict, **kwargs)


inception_encoders = {
    'inceptionresnetv2': {
        'encoder': InceptionResNetV2Encoder,
        'pretrained_settings': pretrained_settings['inceptionresnetv2'],
        'out_shapes': (1536, 1088, 320, 192, 64),
        'params': {
            'num_classes': 1000,
        }

    }
}



import torch
import torch.nn.functional as F
import torch.nn as nn
import torch.utils.model_zoo as model_zoo


def conv_bn(inp, oup, stride, BatchNorm):
    return nn.Sequential(
        nn.Conv2d(inp, oup, 3, stride, 1, bias=False),
        BatchNorm(oup),
        nn.ReLU6(inplace=True)
    )


def fixed_padding(inputs, kernel_size, dilation):
    kernel_size_effective = kernel_size + (kernel_size - 1) * (dilation - 1)
    pad_total = kernel_size_effective - 1
    pad_beg = pad_total // 2
    pad_end = pad_total - pad_beg
    padded_inputs = F.pad(inputs, (pad_beg, pad_end, pad_beg, pad_end))
    return padded_inputs


class InvertedResidual(nn.Module):
    def __init__(self, inp, oup, stride, dilation, expand_ratio, BatchNorm):
        super(InvertedResidual, self).__init__()
        self.stride = stride
        assert stride in [1, 2]

        hidden_dim = round(inp * expand_ratio)
        self.use_res_connect = self.stride == 1 and inp == oup
        self.kernel_size = 3
        self.dilation = dilation

        if expand_ratio == 1:
            self.conv = nn.Sequential(
                # dw
                nn.Conv2d(hidden_dim, hidden_dim, 3, stride, 0, dilation,
                          groups=hidden_dim, bias=False),
                BatchNorm(hidden_dim),
                nn.ReLU6(inplace=True),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, 1, 1, bias=False),
                BatchNorm(oup),
            )
        else:
            self.conv = nn.Sequential(
                # pw
                nn.Conv2d(inp, hidden_dim, 1, 1, 0, 1, bias=False),
                BatchNorm(hidden_dim),
                nn.ReLU6(inplace=True),
                # dw
                nn.Conv2d(hidden_dim, hidden_dim, 3, stride, 0, dilation,
                          groups=hidden_dim, bias=False),
                BatchNorm(hidden_dim),
                nn.ReLU6(inplace=True),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, 1, bias=False),
                BatchNorm(oup),
            )

    def forward(self, x):
        x_pad = fixed_padding(x, self.kernel_size, dilation=self.dilation)
        if self.use_res_connect:
            x = x + self.conv(x_pad)
        else:
            x = self.conv(x_pad)
        return x


class MobileNetV2(nn.Module):
    def __init__(self, output_stride=8, BatchNorm=None, width_mult=1.,
                 pretrained=True):
        super(MobileNetV2, self).__init__()
        block = InvertedResidual
        input_channel = 32
        current_stride = 1
        rate = 1
        interverted_residual_setting = [
            # t, c, n, s
            [1, 16, 1, 1],
            [6, 24, 2, 2],
            [6, 32, 3, 2],
            [6, 64, 4, 2],
            [6, 96, 3, 1],
            [6, 160, 3, 2],
            [6, 320, 1, 1],
        ]

        # building first layer
        input_channel = int(input_channel * width_mult)
        self.features = [conv_bn(3, input_channel, 2, BatchNorm)]
        current_stride *= 2
        # building inverted residual blocks
        for t, c, n, s in interverted_residual_setting:
            if current_stride == output_stride:
                stride = 1
                dilation = rate
                rate *= s
            else:
                stride = s
                dilation = 1
                current_stride *= s
            output_channel = int(c * width_mult)
            for i in range(n):
                if i == 0:
                    self.features.append(
                        block(input_channel, output_channel, stride, dilation,
                              t, BatchNorm))
                else:
                    self.features.append(
                        block(input_channel, output_channel, 1, dilation, t,
                              BatchNorm))
                input_channel = output_channel
        self.features = nn.Sequential(*self.features)
        self._initialize_weights()

        if pretrained:
            self._load_pretrained_model()

        self.low_level_features = self.features[0:4]
        self.high_level_features = self.features[4:]

    def forward(self, x):
        low_level_feat = self.low_level_features(x)
        x = self.high_level_features(low_level_feat)
        return x, low_level_feat

    def _load_pretrained_model(self):
        pretrain_dict = model_zoo.load_url(
            'http://jeff95.me/models/mobilenet_v2-6a65762b.pth')
        model_dict = {}
        state_dict = self.state_dict()
        for k, v in pretrain_dict.items():
            if k in state_dict:
                model_dict[k] = v
        state_dict.update(model_dict)
        self.load_state_dict(state_dict)

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                # n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                # m.weight.data.normal_(0, math.sqrt(2. / n))
                torch.nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()


if __name__ == '__main__':
    input = torch.rand(1, 3, 512, 512)
    model = MobileNetV2(output_stride=16, BatchNorm=nn.BatchNorm2d)
    output, low_level_feat = model(input)
    print(output.size())
    print(low_level_feat.size())



import torch
import torch.nn as nn
import torch.nn.functional as F


class Decoder(nn.Module):
    def __init__(self, num_classes, backbone, BatchNorm):
        super(Decoder, self).__init__()
        if backbone == 'resnet' or backbone == 'drn':
            low_level_inplanes = 256
        elif backbone == 'xception':
            low_level_inplanes = 128
        elif backbone == 'mobilenet':
            low_level_inplanes = 24
        else:
            raise NotImplementedError

        self.conv1 = nn.Conv2d(low_level_inplanes, 48, 1, bias=False)
        self.bn1 = BatchNorm(48)
        self.relu = nn.ReLU()
        self.last_conv = nn.Sequential(
            nn.Conv2d(304, 256, kernel_size=3, stride=1, padding=1,
                      bias=False),
            BatchNorm(256),
            nn.ReLU(),
            nn.Dropout(0.5),
            nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1,
                      bias=False),
            BatchNorm(256),
            nn.ReLU(),
            nn.Dropout(0.1),
            nn.Conv2d(256, num_classes, kernel_size=1, stride=1))
        self._init_weight()

    def forward(self, x, low_level_feat):
        low_level_feat = self.conv1(low_level_feat)
        low_level_feat = self.bn1(low_level_feat)
        low_level_feat = self.relu(low_level_feat)

        x = F.interpolate(x, size=low_level_feat.size()[2:], mode='bilinear',
                          align_corners=True)
        x = torch.cat((x, low_level_feat), dim=1)
        x = self.last_conv(x)

        return x

    def _init_weight(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                torch.nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()


def build_decoder(num_classes, backbone, BatchNorm):
    return Decoder(num_classes, backbone, BatchNorm)



from collections import OrderedDict
import os
from copy import deepcopy

from mlcomp.contrib.search.grid import grid_cells
from mlcomp.db.core import Session
from mlcomp.db.enums import TaskType, DagType, ComponentType
from mlcomp.db.models import Report, Task, Dag, ReportTasks, TaskDependence
from mlcomp.db.providers import TaskProvider, \
    ReportProvider, \
    ReportTasksProvider, \
    ReportLayoutProvider, \
    DagProvider, \
    ProjectProvider
from mlcomp.utils.misc import now
from mlcomp.worker.executors import Executor
from mlcomp.worker.storage import Storage
from mlcomp.utils.io import yaml_dump


class DagStandardBuilder:
    def __init__(
            self,
            session: Session,
            config: dict,
            debug: bool,
            config_text: str = None,
            upload_files: bool = True,
            copy_files_from: int = None,
            config_path: str = None,
            control_reqs: bool = True,
            logger=None,
            component: ComponentType = None,
            grid_cell: dict = None
    ):
        self.session = session
        self.config = config
        self.debug = debug
        self.config_text = config_text
        self.upload_files = upload_files
        self.copy_files_from = copy_files_from
        self.config_path = config_path
        self.control_reqs = control_reqs

        self.info = config['info']
        self.layout_name = self.info.get('layout')

        self.provider = None
        self.report_provider = None
        self.report_tasks_provider = None
        self.report_layout_provider = None
        self.storage = None
        self.dag_provider = None
        self.logger = logger
        self.component = component
        self.grid_cell = grid_cell

        self.project = None
        self.layouts = None
        self.dag = None
        self.dag_report_id = None
        self.created = None
        self.project_provider = None

    def log_info(self, message: str):
        if self.logger:
            self.logger.info(message, self.component)

    def create_providers(self):
        self.log_info('create_providers')

        self.provider = TaskProvider(self.session)
        self.report_provider = ReportProvider(self.session)
        self.report_tasks_provider = ReportTasksProvider(self.session)
        self.report_layout_provider = ReportLayoutProvider(self.session)
        self.project_provider = ProjectProvider(self.session)

        self.storage = Storage(self.session, logger=self.logger,
                               component=self.component)
        self.dag_provider = DagProvider(self.session)

    def load_base(self):
        self.log_info('load_base')

        project = self.project_provider.by_name(self.info['project'])
        if project is None:
            project = self.project_provider.add_project(self.info['project'])

        self.project = project.id
        self.layouts = self.report_layout_provider.all()

    def create_report(self):
        self.log_info('create_report')

        self.dag_report_id = None
        layout_name = self.layout_name
        if layout_name:
            if layout_name not in self.layouts:
                raise Exception(f'Unknown layout = {layout_name}')

            report = Report(
                config=yaml_dump(self.layouts[layout_name]),
                name=self.info['name'],
                project=self.project,
                layout=layout_name
            )
            self.report_provider.add(report)
            self.dag_report_id = report.id

    def create_dag(self):
        self.log_info('create_dag')

        name = self.info['name']
        if self.grid_cell:
            name = f'{name} {self.grid_cell[1]}'

        dag = Dag(
            config=self.config_text or yaml_dump(self.config),
            project=self.project,
            name=name,
            docker_img=self.info.get('docker_img'),
            type=DagType.Standard.value,
            created=now(),
            report=self.dag_report_id
        )

        self.dag = self.dag_provider.add(dag)

    def upload(self):
        self.log_info('upload')

        if self.upload_files:
            folder = os.path.dirname(os.path.abspath(self.config_path))
            if 'expdir' in self.config['info']:
                path = os.path.dirname(os.path.abspath(self.config_path))
                folder = os.path.abspath(
                    os.path.join(path, self.config['info']['expdir'])
                )
            self.storage.upload(folder, self.dag,
                                control_reqs=self.control_reqs)
        elif self.copy_files_from:
            self.storage.copy_from(self.copy_files_from, self.dag)

    def create_task(self, k: str, v: dict, name: str, info: dict,
                    cell: dict = None):
        task_type = TaskType.User.value
        v = deepcopy(v)
        if v.get('task_type') == 'train' or \
                Executor.is_trainable(v['type']):
            task_type = TaskType.Train.value

        gpu = str(v.get('gpu', '0'))
        if '-' not in gpu:
            gpu = int(gpu)
            gpu_max = gpu
        else:
            gpu, gpu_max = map(int, gpu.split('-'))

        if gpu == 0 and gpu_max > 0:
            raise Exception(f'Executor = {k} Gpu_max can"t be>0 when gpu=0')

        task = Task(
            name=name,
            executor=k,
            computer=self.info.get('computer') or v.get('computer'),
            gpu=gpu,
            gpu_max=gpu_max,
            cpu=v.get('cpu', 1),
            memory=v.get('memory', 0.1),
            dag=self.dag.id,
            debug=self.debug,
            steps=int(v.get('steps', '1')),
            type=task_type
        )

        if cell is not None:
            v.update(cell)

        info['executor'] = v
        task.additional_info = yaml_dump(info)
        report = None

        if self.layout_name and task_type == TaskType.Train.value:
            if self.layout_name not in self.layouts:
                raise Exception(f'Unknown report = {v["report"]}')

            report_config = self.layouts[self.layout_name]
            info['report_config'] = report_config

            task.additional_info = yaml_dump(info)
            report = Report(
                config=yaml_dump(report_config),
                name=task.name,
                project=self.project,
                layout=self.layout_name
            )

        return task, report

    def create_tasks(self):
        self.log_info('create_tasks')

        created = OrderedDict()
        executors = self.config['executors']

        tasks = []
        dependencies = []
        reports = []

        while len(created) < len(executors):
            for k, v in executors.items():
                if self.grid_cell:
                    v.update(**self.grid_cell[0])

                valid = True
                if 'depends' in v:
                    depends = v['depends']
                    if not isinstance(depends, list):
                        depends = [depends]

                    for d in depends:
                        if d == k:
                            raise Exception(f'Executor {k} depends on itself')

                        if d not in executors:
                            raise Exception(
                                f'Executor {k} depend on {d} '
                                f'which does not exist'
                            )

                        valid = valid and d in created
                if valid:
                    names = []
                    infos = []
                    task_cells = []
                    if 'grid' in v:
                        grid = v['grid']
                        del v['grid']

                        cells = grid_cells(grid)
                        for i, (cell, cell_name) in enumerate(cells):
                            names.append(cell_name)
                            infos.append({'grid_cell': i})
                            task_cells.append(cell)
                    else:
                        names.append(v.get('name', k))
                        infos.append({})
                        task_cells.append({})

                    k_tasks = []
                    for name, cell, info in zip(names, task_cells, infos):
                        task, report = self.create_task(k, v, name=name,
                                                        info=info, cell=cell)
                        tasks.append(task)
                        k_tasks.append(task)
                        reports.append(report)

                        if 'depends' in v:
                            depends = v['depends']
                            if not isinstance(depends, list):
                                depends = [depends]

                            for d in depends:
                                for dd in created[d]:
                                    dependencies.append((task, dd))
                    created[k] = k_tasks

        not_empty_reports = [r for r in reports if r is not None]
        if len(not_empty_reports) > 0:
            self.provider.bulk_save_objects(not_empty_reports,
                                            return_defaults=True)
            for report, task in zip(reports, tasks):
                if report is not None:
                    task.report = report.id

        self.provider.bulk_save_objects(tasks,
                                        return_defaults=True)

        if len(not_empty_reports) > 0:
            report_tasks = []
            for report, task in zip(reports, tasks):
                if report is not None:
                    report_tasks.append(
                        ReportTasks(report=report.id, task=task.id))
            self.report_tasks_provider.bulk_save_objects(report_tasks)

        dependencies = [
            TaskDependence(task_id=task.id, depend_id=dd.id) for task, dd in
            dependencies
        ]
        self.provider.bulk_save_objects(dependencies)

        for k, v in created.items():
            created[k] = [vv.id for vv in v]
        self.created = created

    def build(self):
        self.create_providers()

        self.load_base()

        self.create_report()

        self.create_dag()

        self.upload()

        self.create_tasks()

        self.log_info('Done')

        return self.created


def dag_standard(
        session: Session,
        config: dict,
        debug: bool,
        config_text: str = None,
        upload_files: bool = True,
        copy_files_from: int = None,
        config_path: str = None,
        control_reqs: bool = True,
        logger=None,
        component: ComponentType = None,
        grid_cell: dict = None
):
    builder = DagStandardBuilder(
        session=session,
        config=config,
        debug=debug,
        config_text=config_text,
        upload_files=upload_files,
        copy_files_from=copy_files_from,
        config_path=config_path,
        control_reqs=control_reqs,
        logger=logger,
        component=component,
        grid_cell=grid_cell
    )
    return builder.build()


__all__ = ['dag_standard']



from .decoder import UnetDecoder
from ..base import EncoderDecoder
from ..encoders import get_encoder


class Unet(EncoderDecoder):
    """Unet_ is a fully convolution neural
    network for image semantic segmentation
    Args:
        encoder_name: name of classification model
        (without last dense layers) used as feature
            extractor to build segmentation model.
        encoder_weights: one of ``None`` (random initialization),
        ``imagenet`` (pre-training on ImageNet).
        decoder_channels: list of numbers of ``Conv2D``
         layer filters in decoder blocks
        decoder_use_batchnorm: if ``True``, ``BatchNormalisation``
         layer between ``Conv2D`` and ``Activation`` layers
            is used.
        classes: a number of classes for output
         (output shape - ``(batch, classes, h, w)``).
        activation: activation function used
        in ``.predict(x)`` method for inference.
            One of [``sigmoid``, ``softmax``, callable, None]
        center: if ``True`` add ``Conv2dReLU``
        block on encoder head (useful for VGG models)
        attention_type: attention module used in decoder of the model
            One of [``None``, ``scse``]
    Returns:
        ``torch.nn.Module``: **Unet**
    .. _Unet:
        https://arxiv.org/pdf/1505.04597
    """

    def __init__(
            self,
            encoder_name='resnet34',
            encoder_weights='imagenet',
            decoder_use_batchnorm=True,
            decoder_channels=(256, 128, 64, 32, 16),
            classes=1,
            activation='sigmoid',
            center=False,  # usefull for VGG models
            attention_type=None
    ):
        encoder = get_encoder(
            encoder_name,
            encoder_weights=encoder_weights
        )

        decoder = UnetDecoder(
            encoder_channels=encoder.out_shapes,
            decoder_channels=decoder_channels,
            final_channels=classes,
            use_batchnorm=decoder_use_batchnorm,
            center=center,
            attention_type=attention_type
        )

        super().__init__(encoder, decoder, activation)

        self.name = 'u-{}'.format(encoder_name)



from torch.nn import CrossEntropyLoss
from torch.nn.functional import nll_loss, log_softmax


class LabelSmoothingCrossEntropy(CrossEntropyLoss):
    def __init__(self, eps: float = 0.1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.eps = eps

    def forward(self, output, target):
        c = output.size()[-1]
        log_preds = log_softmax(output, dim=-1)
        if self.reduction == 'sum':
            loss = -log_preds.sum()
        else:
            loss = -log_preds.sum(dim=-1)
            if self.reduction == 'mean':
                loss = loss.mean()
        nl = nll_loss(log_preds, target, reduction=self.reduction)
        return loss * self.eps / c + (1 - self.eps) * nl


__all__ = ['LabelSmoothingCrossEntropy']



from typing import List
from copy import deepcopy

from mlcomp.db.report_info.f1 import ReportLayoutF1
from mlcomp.db.report_info.img_classify import ReportLayoutImgClassify
from mlcomp.db.report_info.img_segment import ReportLayoutImgSegment
from mlcomp.db.report_info.metric import ReportLayoutMetric
from mlcomp.db.report_info.precision_recall import ReportLayoutPrecisionRecall
from mlcomp.db.report_info.series import ReportLayoutSeries


class ReportLayoutInfo:
    def __init__(self, data: dict):
        assert 'items' in data, 'items is required'
        assert 'metric' in data, 'metric is required'
        assert 'layout' in data, 'layout is required'

        self.data = data
        self.series = self._get_series()
        self.precision_recall = self._get_precision_recall()
        self.metric = self._get_metric()
        self.f1 = self._get_f1()
        self.img_classify = self._get_img_classify()
        self.img_segment = self._get_img_segment()
        self.layout = {'type': 'root', 'items': data.get('layout', [])}
        self._check_layout(self.layout)

    def _check_layout(self, item):
        types = [
            'root', 'panel', 'blank', 'series', 'table', 'img_classify', 'img',
            'img_segment'
        ]
        assert item.get('type') in types, f'Unknown item type = {item["type"]}'

        fields = {
            'root': [('items', False)],
            'panel': [
                'title', ('parent_cols', False), ('cols', False),
                ('row_height', False), ('rows', False), ('items', False),
                ('expanded', False), ('table', False)
            ],
            'blank': [('cols', False), ('rows', False)],
            'series': [
                ('multi', False), ('group', False), 'source', ('cols', False),
                ('rows', False)
            ],
            'table': [
                'source',
                ('cols', False),
                ('rows', False),
            ],
            'img_classify': [
                'source', ('attrs', False), ('cols', False), ('rows', False)
            ],
            'img_segment': [
                'source', ('attrs', False), ('cols', False), ('rows', False),
                ('max_width', False), ('max_height', False)
            ],
            'img': ['source', ('cols', False), ('rows', False)]
        }
        keys = set(item.keys()) - {'type'}
        for f in fields[item['type']]:
            req = True
            if type(f) == tuple:
                f, req = f
            if req and f not in item:
                raise Exception(f'Type {item["type"]} must contain field {f}')
            if f in keys:
                keys.remove(f)
        assert len(keys) == 0, \
            f'Unknown fields {keys} for type = {item["type"]}'

        if 'items' in item:
            for item in item['items']:
                self._check_layout(item)

    def has_classification(self):
        return len(self.precision_recall) > 0

    def _by_type(self, t: str, c):
        return [
            c.from_dict(k, v) for k, v in self.data['items'].items()
            if v.get('type') == t
        ]

    def _get_img_classify(self) -> List[ReportLayoutImgClassify]:
        return self._by_type('img_classify', ReportLayoutImgClassify)

    def _get_img_segment(self) -> List[ReportLayoutImgSegment]:
        return self._by_type('img_segment', ReportLayoutImgSegment)

    def _get_f1(self) -> List[ReportLayoutF1]:
        return self._by_type('f1', ReportLayoutF1)

    def _get_series(self) -> List[ReportLayoutSeries]:
        return self._by_type('series', ReportLayoutSeries)

    def _get_precision_recall(self) -> List[ReportLayoutPrecisionRecall]:
        return self._by_type('precision_recall', ReportLayoutPrecisionRecall)

    def _get_metric(self) -> ReportLayoutMetric:
        return ReportLayoutMetric.from_dict(self.data['metric'])

    @classmethod
    def union_layouts(cls, name: str, layouts: dict, return_dict: bool = True):
        assert name in layouts, f'Layout {name} is not in the collection'

        layout = deepcopy(layouts[name])
        r = dict()
        if layout.get('extend'):
            assert layout['extend'] in layouts, \
                f'Layout for extending = {layout["extend"]}' \
                f' is not in the collection'
            r = cls.union_layouts(layout['extend'], layouts)

        if 'metric' in layout:
            r['metric'] = layout['metric']

        if 'items' in layout:
            r['items'] = r.get('items', dict())
            r['items'].update(layout['items'])

        if 'layout' in layout:
            r['layout'] = r.get('layout', []) + layout['layout']

        if return_dict:
            return r

        return ReportLayoutInfo(r)


__all__ = ['ReportLayoutInfo']



from mlcomp.db.models import Log, Step, Task, Computer
from mlcomp.db.core import PaginatorOptions
from mlcomp.db.enums import ComponentType
from mlcomp.db.providers.base import BaseDataProvider
from mlcomp.utils.misc import log_name, to_snake


class LogProvider(BaseDataProvider):
    model = Log

    def get(self, filter: dict, options: PaginatorOptions):
        query = self.query(Log, Step, Task). \
            join(Step, Step.id == Log.step, isouter=True). \
            join(Task, Task.id == Log.task, isouter=True)

        if filter.get('message'):
            query = query.filter(Log.message.contains(filter['message']))

        if filter.get('dag'):
            query = query.filter(Task.dag == filter['dag'])

        if filter.get('task'):
            child_tasks = self.query(Task.id
                                     ).filter(Task.parent == filter['task']
                                              ).all()
            child_tasks = [c[0] for c in child_tasks]
            child_tasks.append(filter['task'])

            query = query.filter(Task.id.in_(child_tasks))

        if len(filter.get('components', [])) > 0:
            query = query.filter(Log.component.in_(filter['components']))

        if filter.get('computer'):
            query = query.filter(Computer.name == filter['computer'])

        if len(filter.get('levels', [])) > 0:
            query = query.filter(Log.level.in_(filter['levels']))

        if filter.get('task_name'):
            query = query.filter(Task.name.like(f'%{filter["task_name"]}%'))

        if filter.get('step_name'):
            query = query.filter(Step.name.like(f'%{filter["step_name"]}%'))

        if filter.get('step'):
            query = query.filter(Step.id == filter['step'])

        total = query.count()
        data = []
        for log, step, task in self.paginator(query, options):
            item = {
                'id': log.id,
                'message': log.message.split('\n'),
                'module': log.module,
                'line': log.line,
                'time': self.serializer.serialize_datetime(log.time),
                'level': log_name(log.level),
                'component': to_snake(ComponentType(log.component).name),
                'computer': log.computer,
                'step': self.to_dict(step) if step else None,
                'task': self.to_dict(task, rules=('-additional_info',))
                if task else None
            }
            data.append(item)

        return {'total': total, 'data': data}

    def last(self, count: int, dag: int = None, task: int = None,
             levels=None, components=None):
        query = self.query(Log, Task.id).outerjoin(Task)
        if dag is not None:
            query = query.filter(Task.dag == dag)
        if task is not None:
            query = query.filter(Task.id == task)
        if levels is not None:
            query = query.filter(Log.level.in_(levels))
        if components is not None:
            query = query.filter(Log.component.in_(components))
        return query.order_by(Log.id.desc()).limit(count).all()


__all__ = ['LogProvider']



import os
import pickle
import random
import warnings
from collections import defaultdict
from numbers import Number
from os.path import join
from typing import Callable, Dict, List

import cv2
import numpy as np
import pandas as pd
from torch.utils.data import Dataset
from torchvision.datasets.video_utils import VideoClips

warnings.filterwarnings('ignore')


class VideoClipsFolder:
    def __init__(self, video_paths: List[str], clip_length_in_frames: int,
                 frames_between_clips: int,
                 _precomputed_metadata: dict = None):
        self.video_paths = video_paths
        self.clip_length_in_frames = clip_length_in_frames
        self.frames_between_clips = frames_between_clips
        self.cumulative_sizes = []
        self.clips = []
        self.metadata = _precomputed_metadata
        self.compute_clips()

    # noinspection PyTypeChecker
    def compute_clips(self):
        if self.metadata is not None:
            self.clips = self.metadata['clips']
            self.cumulative_sizes = self.metadata[
                'cumulative_sizes']
            self.video_paths = self.metadata['video_paths']
            return

        clips_size = 0
        for video_index, folder in enumerate(self.video_paths):
            files = sorted(os.listdir(folder))
            assert len(files) >= self.clip_length_in_frames, \
                f'folder = {folder} has only {len(files)} files'

            for i in range(0, len(files) - self.clip_length_in_frames,
                           self.frames_between_clips):
                clips_size += 1
                self.clips.append({
                    'video_index': video_index,
                    'min_index': i,
                    'max_index': i + self.clip_length_in_frames
                })

            self.cumulative_sizes.append(clips_size)

        self.metadata = {
            'clips': self.clips,
            'cumulative_sizes': self.cumulative_sizes,
            'video_paths': self.video_paths
        }

    def get_clip(self, index: int):
        clip = self.clips[index]
        files = sorted(os.listdir(self.video_paths[clip['video_index']]))
        imgs = []
        for index in range(clip['min_index'], clip['max_index']):
            file = join(self.video_paths[clip['video_index']], files[index])
            img = cv2.imread(file)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            imgs.append(img)
        return imgs, None, None, clip['video_index']


class VideoDataset(Dataset):
    def __init__(
            self,
            *,
            video_folder: str,
            fold_csv: str = None,
            fold: int = None,
            is_test: bool = False,
            clip_length_in_frames: int = 1,
            frames_between_clips: int = 1,
            num_classes=2,
            max_count=None,
            transforms=None,
            postprocess_func: Callable[[Dict], Dict] = None,
            clips_per_video: int = 1,
            metadata_path: str = None
    ):
        self.video_folder = video_folder

        if fold_csv:
            df = pd.read_csv(fold_csv)
            if fold is not None:
                if is_test:
                    self.data = df[df['fold'] == fold]
                else:
                    self.data = df[df['fold'] != fold]
            else:
                self.data = df
        else:
            self.data = pd.DataFrame(
                {'video': os.listdir(video_folder)}).sort_values(by='video')

        self.data = self.data.sample(frac=1)
        self.data = self.data.to_dict(orient='row')
        if max_count is not None:
            self.apply_max_count(max_count)

        for row in self.data:
            self.preprocess_row(row)

        self.transforms = transforms
        self.num_classes = num_classes
        self.postprocess_func = postprocess_func
        self.clip_length_in_frames = clip_length_in_frames
        self.frames_between_clips = frames_between_clips
        self.clips_per_video = clips_per_video
        self.metadata_path = metadata_path

        self.video_paths = [row['video'] for row in self.data]
        self.scan_mode = any(os.path.isdir(v) for v in self.video_paths)
        self.clips = self.create_clips()

    def create_clips(self):
        precomputed_metadata = None
        if self.metadata_path:
            if os.path.exists(self.metadata_path):
                precomputed_metadata = pickle.load(
                    open(self.metadata_path, 'rb'))
                if precomputed_metadata is not None:
                    if len(precomputed_metadata['video_paths']) != len(
                            self.video_paths):
                        precomputed_metadata = None

                    if precomputed_metadata is not None and any(
                            [p1 != p2 for p1, p2 in zip(self.video_paths,
                                                        precomputed_metadata[
                                                            'video_paths'])]):
                        precomputed_metadata = None

        if precomputed_metadata is None:
            print('computing metadata')

        cls = VideoClipsFolder if self.scan_mode else VideoClips
        clips = cls(
            video_paths=self.video_paths,
            clip_length_in_frames=self.clip_length_in_frames,
            frames_between_clips=self.frames_between_clips,
            _precomputed_metadata=precomputed_metadata
        )

        if precomputed_metadata is None and self.metadata_path:
            folder = os.path.dirname(self.metadata_path)
            os.makedirs(folder, exist_ok=True)
            pickle.dump(clips.metadata, open(self.metadata_path, 'wb'))

        return clips

    def apply_max_count(self, max_count):
        if isinstance(max_count, Number):
            self.data = self.data[:max_count]
        else:
            data = defaultdict(list)
            for row in self.data:
                data[row['label']].append(row)
            min_index = np.argmin(max_count)
            min_count = len(data[min_index])
            for k, v in data.items():
                count = int(min_count * (max_count[k] / max_count[min_index]))
                data[k] = data[k][:count]

            self.data = [v for i in range(len(data)) for v in data[i]]

    def preprocess_row(self, row: dict):
        row['video'] = join(self.video_folder, row['video'])

    def __len__(self):
        return len(self.clips.video_paths) * self.clips_per_video

    def __getitem__(self, index):
        video_index = index // self.clips_per_video
        max_clip_index = self.clips.cumulative_sizes[video_index]
        min_clip_index = self.clips.cumulative_sizes[
            video_index - 1] if video_index > 0 else 0

        clip_index = np.random.randint(min_clip_index, max_clip_index)
        video, audio, info, _ = self.clips.get_clip(clip_index)
        row = self.data[video_index]

        if self.transforms:
            seed = random.randint(0, 10 ** 6)
            frames = []
            if not self.scan_mode:
                # noinspection PyUnresolvedReferences
                video = video.numpy()

            for v in video:
                random.seed(seed)
                frames.append(self.transforms(image=v)['image'][None])

            video = np.vstack(frames)

        res = {
            'features': video,
            'video_file': row['video']
        }
        if 'label' in row:
            res['targets'] = row['label']

        if self.postprocess_func:
            res = self.postprocess_func(res)
        return res


__all__ = ['VideoDataset']



# flake8: noqa
# noinspection PyUnresolvedReferences
from mlcomp.utils.tests import session
from mlcomp.db.core import Session
from mlcomp.db.providers import ProjectProvider


class TestProject(object):

    def _configure(self, session):
        provider = ProjectProvider(session)
        provider.add_project(name='test')
        return provider

    def test_add(self, session: Session):
        provider = ProjectProvider(session)
        project = provider.add_project(name='test')
        assert provider.by_id(project.id)

    def test_by_name(self, session: Session):
        provider = self._configure(session)
        project = provider.by_name('test')
        assert project is not None

    def test_get(self, session: Session):
        provider = self._configure(session)
        res = provider.get()
        assert len(res['data']) == 1
        assert res['total'] == 1



from mlcomp.db.core import Session
from mlcomp.db.enums import ComponentType
from mlcomp.db.models import Task, Step
from mlcomp.db.providers import LogProvider, StepProvider, TaskProvider
from mlcomp.utils.misc import now


class StepWrap:
    def __init__(
            self, session: Session, logger, logger_db, task: Task,
            task_provider: TaskProvider
    ):
        self.log_provider = LogProvider(session)
        self.step_provider = StepProvider(session)
        self.task_provider = task_provider
        self.task = task
        self.children = []
        self.step = None
        self.logger = logger
        self.logger_db = logger_db

    @property
    def id(self):
        return self.step.id

    def enter(self):
        task = self.task if not self.task.parent else self.task_provider.by_id(
            self.task.parent
        )
        self.children = self.step_provider.unfinished(task.id)
        if len(self.children) == 0:
            self.step = self.start(0, 'main', 0)
        else:
            self.step = self.children[-1]

    def _finish(self):
        if len(self.children) == 0:
            return
        step = self.children.pop()
        step.finished = now()
        self.step_provider.update()
        self.step = self.children[-1] if len(self.children) > 0 else step

        self.debug('End of the step')

    def finish(self):
        while len(self.children) > 0:
            self._finish()

    def start(self, level: int, name: str = None, index: int = None):
        task = self.task if not self.task.parent else self.task_provider.by_id(
            self.task.parent
        )

        if index is None and task.current_step:
            parts = task.current_step.split('.')
            if len(parts) >= level:
                index = int(parts[level - 1])

        if self.step and index == self.step.index and self.step.level == level:
            return

        if self.step is not None:
            diff = level - self.step.level
            assert level > 0, 'level must be positive'
            assert diff <= 1, \
                f'Level {level} can not be started after {self.step.level}'

            if diff <= 0:
                for _ in range(abs(diff) + 1):
                    self._finish()

        step = Step(
            level=level,
            name=name or '',
            started=now(),
            task=task.id,
            index=index or 0
        )
        self.step_provider.add(step)
        self.children.append(step)
        self.step = step

        task.current_step = '.'.join(
            [
                str(c.index + 1)
                for c in self.children[1:]
            ]
        )
        self.task_provider.commit()

        self.debug('Begin of the step')

        return step

    def end(self, level: int):
        diff = level - self.step.level
        assert diff <= 0, 'you can end only the same step or lower'
        for i in range(abs(diff) + 1):
            self._finish()

    def debug(self, message: str, db: bool = False):
        logger = self.logger_db if db else self.logger
        logger.debug(
            message, ComponentType.Worker, self.task.computer_assigned,
            self.task.id, self.step.id
        )

    def info(self, message: str, db: bool = False):
        logger = self.logger_db if db else self.logger
        logger.info(
            message, ComponentType.Worker, self.task.computer_assigned,
            self.task.id, self.step.id
        )

    def warning(self, message: str, db: bool = False):
        logger = self.logger_db if db else self.logger
        logger.warning(
            message, ComponentType.Worker, self.task.computer_assigned,
            self.task.id, self.step.id
        )

    def error(self, message: str, db: bool = False):
        logger = self.logger_db if db else self.logger
        logger.error(
            message, ComponentType.Worker, self.task.computer_assigned,
            self.task.id, self.step.id
        )


__all__ = ['StepWrap']



import torch.nn as nn


class LambdaLayer(nn.Module):
    def __init__(self, lambd):
        super(LambdaLayer, self).__init__()
        self.lambd = lambd

    def forward(self, x):
        return self.lambd(x)


__all__ = ['LambdaLayer']



import torch.nn as nn
from torchvision.models.vgg import VGG
from torchvision.models.vgg import make_layers
from pretrainedmodels.models.torchvision_models import pretrained_settings

cfg = {
    'A': [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'B': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512,
          'M'],
    'D': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M',
          512, 512, 512, 'M'],
    'E': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512,
          512, 'M', 512, 512, 512, 512, 'M'],
}


class VGGEncoder(VGG):

    def __init__(self, config, batch_norm=False, *args, **kwargs):
        super().__init__(
            make_layers(config, batch_norm=batch_norm),
            *args,
            **kwargs
        )
        self.pretrained = False
        del self.classifier

    def forward(self, x):
        features = []
        for module in self.features:
            if isinstance(module, nn.MaxPool2d):
                features.append(x)
            x = module(x)
        features.append(x)

        features = features[1:]
        features = features[::-1]
        return features

    def load_state_dict(self, state_dict, **kwargs):
        keys = list(state_dict.keys())
        for k in keys:
            if k.startswith('classifier'):
                state_dict.pop(k)
        super().load_state_dict(state_dict, **kwargs)


vgg_encoders = {

    'vgg11': {
        'encoder': VGGEncoder,
        'out_shapes': (512, 512, 512, 256, 128),
        'pretrained_settings': pretrained_settings['vgg11'],
        'params': {
            'config': cfg['A'],
            'batch_norm': False,
        },
    },

    'vgg11_bn': {
        'encoder': VGGEncoder,
        'out_shapes': (512, 512, 512, 256, 128),
        'pretrained_settings': pretrained_settings['vgg11_bn'],
        'params': {
            'config': cfg['A'],
            'batch_norm': True,
        },
    },

    'vgg13': {
        'encoder': VGGEncoder,
        'out_shapes': (512, 512, 512, 256, 128),
        'pretrained_settings': pretrained_settings['vgg13'],
        'params': {
            'config': cfg['B'],
            'batch_norm': False,
        },
    },

    'vgg13_bn': {
        'encoder': VGGEncoder,
        'out_shapes': (512, 512, 512, 256, 128),
        'pretrained_settings': pretrained_settings['vgg13_bn'],
        'params': {
            'config': cfg['B'],
            'batch_norm': True,
        },
    },

    'vgg16': {
        'encoder': VGGEncoder,
        'out_shapes': (512, 512, 512, 256, 128),
        'pretrained_settings': pretrained_settings['vgg16'],
        'params': {
            'config': cfg['D'],
            'batch_norm': False,
        },
    },

    'vgg16_bn': {
        'encoder': VGGEncoder,
        'out_shapes': (512, 512, 512, 256, 128),
        'pretrained_settings': pretrained_settings['vgg16_bn'],
        'params': {
            'config': cfg['D'],
            'batch_norm': True,
        },
    },

    'vgg19': {
        'encoder': VGGEncoder,
        'out_shapes': (512, 512, 512, 256, 128),
        'pretrained_settings': pretrained_settings['vgg19'],
        'params': {
            'config': cfg['E'],
            'batch_norm': False,
        },
    },

    'vgg19_bn': {
        'encoder': VGGEncoder,
        'out_shapes': (512, 512, 512, 256, 128),
        'pretrained_settings': pretrained_settings['vgg19_bn'],
        'params': {
            'config': cfg['E'],
            'batch_norm': True,
        },
    },
}



import torch.nn as nn


class Conv2dReLU(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, padding=0,
                 stride=1, use_batchnorm=True, **batchnorm_params):
        super().__init__()

        layers = [
            nn.Conv2d(in_channels, out_channels, kernel_size,
                      stride=stride, padding=padding,
                      bias=not (use_batchnorm)),
            nn.ReLU(inplace=True),
        ]

        if use_batchnorm:
            layers.insert(1, nn.BatchNorm2d(out_channels, **batchnorm_params))

        self.block = nn.Sequential(*layers)

    def forward(self, x):
        return self.block(x)


class SCSEModule(nn.Module):
    def __init__(self, ch, re=16):
        super().__init__()
        self.cSE = nn.Sequential(nn.AdaptiveAvgPool2d(1),
                                 nn.Conv2d(ch, ch // re, 1),
                                 nn.ReLU(inplace=True),
                                 nn.Conv2d(ch // re, ch, 1),
                                 nn.Sigmoid()
                                 )
        self.sSE = nn.Sequential(nn.Conv2d(ch, ch, 1),
                                 nn.Sigmoid())

    def forward(self, x):
        return x * self.cSE(x) + x * self.sSE(x)



import sqlalchemy as sa

from mlcomp.db.models.base import Base


class Project(Base):
    __tablename__ = 'project'

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String, nullable=False)
    class_names = sa.Column(sa.String, nullable=False)
    sync_folders = sa.Column(sa.String, nullable=False)
    ignore_folders = sa.Column(sa.String, nullable=False)


__all__ = ['Project']



import torch


def flip(x, dim):
    indices = [slice(None)] * x.dim()
    indices[dim] = torch.arange(x.size(dim) - 1, -1, -1,
                                dtype=torch.long, device=x.device)
    return x[tuple(indices)]


__all__ = ['flip']



import importlib
import sys
from os import getcwd
from os.path import join

from mlcomp.worker.executors import Executor


@Executor.register
class Click(Executor):
    def __init__(self, module: str, command: str = None, **kwargs):
        super().__init__(**kwargs)

        self.module = module
        self.command = command

    def work(self):
        self.info('click. creating module')
        spec = importlib.util.spec_from_file_location(
            self.module,
            join(getcwd(), self.module+'.py')
        )
        m = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(m)
        self.info('click. module created')

        command = getattr(m, self.command)
        self.info('click. command get')

        m.tqdm = self.tqdm

        self.info('click. tqdm set')

        for p in command.params:
            if p.name in self.kwargs:
                p.default = self.kwargs[p.name]

        sys.argv = sys.argv[:1]

        self.info('click. setup finished. executing command')

        stdout = sys.stdout
        sys.stdout = self
        self.info('click. stdout set')

        command(standalone_mode=False)

        sys.stdout = stdout

        self.info('click. command finished')


__all__ = ['Click']



class ReportLayoutMetric:
    def __init__(self, name: str, minimize: bool):
        self.name = name
        self.minimize = minimize

    @staticmethod
    def from_dict(data: dict):
        name = data.pop('name')
        minimize = data.pop('minimize')
        assert len(data) == 0, f'Unknown parameter in ' \
            f'report.metric={data.popitem()}'
        return ReportLayoutMetric(name, minimize)

    def serialize(self):
        return {'minimize': self.minimize, 'name': self.name}


__all__ = ['ReportLayoutMetric']



import sqlalchemy as sa
from sqlalchemy import ForeignKey

from mlcomp.db.models.base import Base


class DagStorage(Base):
    __tablename__ = 'dag_storage'

    id = sa.Column(sa.Integer, primary_key=True)
    dag = sa.Column(sa.Integer, ForeignKey('dag.id'))
    file = sa.Column(sa.Integer, ForeignKey('file.id'))
    path = sa.Column(sa.String)
    is_dir = sa.Column(sa.Boolean)


class DagLibrary(Base):
    __tablename__ = 'dag_library'

    id = sa.Column(sa.Integer, primary_key=True)
    dag = sa.Column(sa.Integer, ForeignKey('dag.id'))
    library = sa.Column(sa.String)
    version = sa.Column(sa.String)


__all__ = ['DagStorage', 'DagLibrary']



import numpy as np

from torch.utils.data import DistributedSampler


class DistributedSamplerIndices(DistributedSampler):
    def __init__(self, sampler, *args, **kwargs):
        super().__init__(sampler, num_replicas=None, rank=None, shuffle=True)
        self.sampler = sampler

    def get_indices(self):
        return list(self.sampler.__iter__())

    def __iter__(self):
        # deterministically shuffle based on epoch
        np.random.seed(self.epoch)

        indices = self.get_indices()

        if self.shuffle:
            np.random.shuffle(indices)

        # add extra samples to make it evenly divisible
        indices += indices[:(self.total_size - len(indices))]
        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank:self.total_size:self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices)


__all__ = ['DistributedSamplerIndices']



from collections import OrderedDict

from mlcomp.db.report_info.item import ReportLayoutItem


class ReportLayoutSeries(ReportLayoutItem):
    def __init__(self, name: str, key: str):
        super().__init__(name)

        self.key = key

    @classmethod
    def from_dict(cls, name: str, value: OrderedDict):
        assert 'key' in value, f'report.series={name}. key is required'
        value.pop('type')
        key = value.pop('key')

        assert len(value) == 0, f'Unknown parameter in ' \
            f'report.series={name}: {value.popitem()}'
        return cls(name, key)


__all__ = ['ReportLayoutSeries']
